from __future__ import annotations

import os
from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


def dependency_lock_click_options(fn):
    """Options shared by ``hatch dep lock`` and top-level ``hatch lock``."""
    decorators = [
        click.option("--upgrade", "-U", is_flag=True, help="Upgrade all packages"),
        click.option("--upgrade-package", "-P", multiple=True, help="Upgrade specific package(s)"),
        click.option(
            "--export", "export_path", type=click.Path(), default=None, help="Export lockfile to a custom path"
        ),
        click.option(
            "--export-all",
            "export_all_path",
            type=click.Path(),
            default=None,
            help="Export lockfiles for all environments to a directory",
        ),
        click.option("--check", is_flag=True, help="Check if lockfile is up-to-date"),
    ]
    for d in reversed(decorators):
        fn = d(fn)
    return fn


def run_dep_lock(
    app: Application,
    *,
    upgrade: bool,
    upgrade_package: tuple[str, ...],
    export_path: str | None,
    export_all_path: str | None,
    check: bool,
) -> None:
    """Lock workflow for the active ``HATCH_ENV`` (``hatch -e``), with ``--export-all`` matching ``hatch env lock``."""
    if export_path and export_all_path:
        app.abort("Cannot use both `--export` and `--export-all`")

    if export_all_path:
        env_names = [*app.project.config.envs, *app.project.config.internal_envs]
        explicit_env = None
        skip_incompatible = True
    else:
        env_names = app.project.expand_environments(app.env)
        if not env_names:
            app.abort(f"Environment `{app.env}` is not defined by project config")
        explicit_env = app.env
        skip_incompatible = app.env in app.project.config.matrices

    run_lock_workflow(
        app,
        env_names,
        upgrade=upgrade,
        upgrade_package=upgrade_package,
        export_path=export_path,
        export_all_path=export_all_path,
        check=check,
        explicit_env=explicit_env,
        skip_incompatible_envs=skip_incompatible,
    )


def run_lock_workflow(
    app: Application,
    env_names: list[str],
    *,
    upgrade: bool,
    upgrade_package: tuple[str, ...],
    export_path: str | None,
    export_all_path: str | None,
    check: bool,
    explicit_env: str | None,
    skip_incompatible_envs: bool,
) -> None:
    """Shared implementation for ``hatch env lock`` and ``hatch dep lock``."""
    from hatch.config.constants import AppEnvVars
    from hatch.env.lock import (
        LockerNotFoundError,
        LockerUnsupportedError,
        environment_has_lock_inputs,
        generate_lockfile,
        lockfile_in_sync,
        merge_environment_lock_inputs,
        resolve_lockfile_path,
    )
    from hatch.utils.fs import Path

    incompatible: dict[str, str] = {}
    lockfile_groups: dict[Path, list[str]] = {}

    for env in env_names:
        environment = app.project.get_environment(env)

        try:
            environment.check_compatibility()
        except Exception as e:  # noqa: BLE001
            if skip_incompatible_envs:
                incompatible[env] = str(e)
                continue

            app.abort(f"Environment `{env}` is incompatible: {e}")

        if explicit_env is not None and not environment.locked and not export_path and not export_all_path:
            app.abort(
                f"Environment `{explicit_env}` is not configured with `locked = true`. "
                f"Use `--export <path>` to generate a lockfile, "
                f"or set `locked = true` on the environment (or `lock-envs = true` globally)."
            )

        if export_path:
            output_path = Path(export_path)
        elif export_all_path:
            safe_name = environment.name.replace(".", "-")
            filename = "pylock.toml" if environment.name == "default" else f"pylock.{safe_name}.toml"
            output_path = Path(export_all_path) / filename
        else:
            output_path = resolve_lockfile_path(environment)

        if check:
            if not output_path.is_file():
                app.abort(f"Lockfile does not exist: {output_path}")
        elif not environment_has_lock_inputs(environment):
            app.display_warning(f"Environment `{env}` has no dependencies to lock")
            continue

        lockfile_groups.setdefault(output_path, []).append(env)

    for output_path, envs in lockfile_groups.items():
        environment, merged_deps, merged_extras, merged_groups, display_name = merge_environment_lock_inputs(
            app.project, envs, app.abort
        )

        try:
            if check:
                with app.status(f"Checking lockfile for: {display_name}"):
                    app.project.prepare_environment(environment, keep_env=bool(os.environ.get(AppEnvVars.KEEP_ENV)))
                    if not lockfile_in_sync(
                        environment,
                        output_path,
                        upgrade=upgrade,
                        upgrade_packages=upgrade_package,
                        deps_override=merged_deps,
                        lock_extras=merged_extras,
                        lock_groups=merged_groups,
                    ):
                        app.abort(f"Lockfile is not up to date: {output_path}")
                app.display(f"Lockfile is up to date: {output_path}")
                continue

            with app.status(f"Locking environment: {display_name}"):
                app.project.prepare_environment(environment, keep_env=bool(os.environ.get(AppEnvVars.KEEP_ENV)))
                generate_lockfile(
                    environment,
                    output_path,
                    upgrade=upgrade,
                    upgrade_packages=upgrade_package,
                    deps_override=merged_deps,
                    lock_extras=merged_extras,
                    lock_groups=merged_groups,
                )

            app.display(f"Wrote lockfile: {output_path}")
        except LockerNotFoundError as e:
            app.abort(str(e))
        except LockerUnsupportedError as e:
            app.abort(str(e))

    if incompatible:
        num_incompatible = len(incompatible)
        app.display_warning(
            f"Skipped {num_incompatible} incompatible environment{'s' if num_incompatible > 1 else ''}:"
        )
        for env, reason in incompatible.items():
            app.display_warning(f"{env} -> {reason}")


@click.command(short_help="Generate lockfiles for environments")
@click.argument("env_name", required=False, default=None)
@click.option("--upgrade", "-U", is_flag=True, help="Upgrade all packages")
@click.option("--upgrade-package", "-P", multiple=True, help="Upgrade specific package(s)")
@click.option("--export", "export_path", type=click.Path(), default=None, help="Export lockfile to a custom path")
@click.option(
    "--export-all",
    "export_all_path",
    type=click.Path(),
    default=None,
    help="Export lockfiles for all environments to a directory",
)
@click.option("--check", is_flag=True, help="Check if lockfile is up-to-date")
@click.pass_obj
def lock(
    app: Application,
    *,
    env_name: str | None,
    upgrade: bool,
    upgrade_package: tuple[str, ...],
    export_path: str | None,
    export_all_path: str | None,
    check: bool,
):
    """Generate lockfiles for environments.

    When called without arguments, locks all environments that have `locked = true`
    in their configuration. When called with ENV_NAME, locks that specific environment.
    """
    app.ensure_environment_plugin_dependencies()

    if export_path and export_all_path:
        app.abort("Cannot use both `--export` and `--export-all`")

    if export_all_path:
        env_names = [*app.project.config.envs, *app.project.config.internal_envs]
        explicit_env = None
        skip_incompatible = True
    elif env_name is not None:
        env_names = app.project.expand_environments(env_name)
        if not env_names:
            app.abort(f"Environment `{env_name}` is not defined by project config")
        explicit_env = env_name
        skip_incompatible = env_name in app.project.config.matrices
    else:
        env_names = []
        for name in app.project.config.envs:
            environment = app.project.get_environment(name)
            if environment.locked:
                env_names.append(name)
        for name in app.project.config.internal_envs:
            environment = app.project.get_environment(name)
            if environment.locked:
                env_names.append(name)

        if not env_names:
            app.abort("No environments are configured with `locked = true`")
        explicit_env = None
        skip_incompatible = True

    run_lock_workflow(
        app,
        env_names,
        upgrade=upgrade,
        upgrade_package=upgrade_package,
        export_path=export_path,
        export_all_path=export_all_path,
        check=check,
        explicit_env=explicit_env,
        skip_incompatible_envs=skip_incompatible,
    )



from __future__ import annotations

import json
import os
import shutil
import subprocess
import sys
from tempfile import TemporaryDirectory

RUNNER: dict = {}


def main() -> int:
    project_root: str = RUNNER["project_root"]
    output_dir: str = RUNNER["output_dir"]
    hook: str = RUNNER["hook"]
    kwargs: dict[str, str] = RUNNER["kwargs"]
    backend: str = RUNNER["backend"]
    backend_path: str = RUNNER["backend_path"]
    hook_caller_script: str = RUNNER["hook_caller_script"]

    with TemporaryDirectory() as d:
        temp_dir = os.path.realpath(d)
        control_dir = os.path.join(temp_dir, "control")
        os.mkdir(control_dir)
        input_file = os.path.join(control_dir, "input.json")
        output_file = os.path.join(control_dir, "output.json")

        env_vars = dict(os.environ)
        env_vars["_PYPROJECT_HOOKS_BUILD_BACKEND"] = backend
        if backend_path:
            env_vars["_PYPROJECT_HOOKS_BACKEND_PATH"] = backend_path

        if "work_dir" in kwargs:
            work_dir = os.path.join(temp_dir, "work")
            os.mkdir(work_dir)
            kwargs[kwargs.pop("work_dir")] = work_dir
        else:
            work_dir = ""

        with open(input_file, "w", encoding="utf-8") as f:
            f.write(json.dumps({"kwargs": kwargs}))

        script_path = os.path.join(temp_dir, "script.py")
        with open(script_path, "w", encoding="utf-8") as f:
            f.write(hook_caller_script)

        process = subprocess.run(
            [sys.executable, script_path, hook, str(control_dir)],
            cwd=project_root,
            env=env_vars,
            check=False,
        )
        if process.returncode:
            return process.returncode

        with open(output_file, encoding="utf-8") as f:
            output = json.loads(f.read())

        if output.get("no_backend", False):
            sys.stderr.write(f"{output['traceback']}\n{output['backend_error']}\n")
            return 1

        if output.get("unsupported", False):
            sys.stderr.write(output["traceback"])
            return 1

        if output.get("hook_missing", False):
            sys.stderr.write(f"Build backend API `{backend}` is missing hook: {output['missing_hook_name']}\n")
            return 1

        shutil.move(output_file, output_dir)
        if work_dir:
            shutil.move(work_dir, output_dir)

    return 0


if __name__ == "__main__":
    code = main()
    sys.stderr.flush()
    os._exit(code)



from __future__ import annotations

import os
import platform
import sys
from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, Literal

from hatch.config.constants import PythonEnvVars
from hatch.errors import PythonDistributionResolutionError, PythonDistributionUnknownError
from hatch.python.distributions import DISTRIBUTIONS, ORDERED_DISTRIBUTIONS

if TYPE_CHECKING:
    from packaging.version import Version

    from hatch.utils.fs import Path


# Use an artificially high epoch to ensure that custom distributions are always considered newer
CUSTOM_DISTRIBUTION_VERSION_EPOCH = 100


def custom_env_var(prefix: str, name: str) -> str:
    return f"{prefix}{name.upper().replace('.', '_')}"


def get_custom_source(name: str) -> str | None:
    return os.environ.get(custom_env_var(PythonEnvVars.CUSTOM_SOURCE_PREFIX, name))


def get_custom_version(name: str) -> str | None:
    return os.environ.get(custom_env_var(PythonEnvVars.CUSTOM_VERSION_PREFIX, name))


def get_custom_path(name: str) -> str | None:
    return os.environ.get(custom_env_var(PythonEnvVars.CUSTOM_PATH_PREFIX, name))


class Distribution(ABC):
    def __init__(self, name: str, source: str) -> None:
        self.__name = name
        self.__source = source

    @property
    def name(self) -> str:
        return self.__name

    @cached_property
    def source(self) -> str:
        return self.__source if (custom_source := get_custom_source(self.name)) is None else custom_source

    @cached_property
    def archive_name(self) -> str:
        return self.source.rsplit("/", 1)[-1]

    def unpack(self, archive: Path, directory: Path) -> None:
        if self.source.endswith(".zip"):
            import zipfile

            with zipfile.ZipFile(archive, "r") as zf:
                zf.extractall(directory)
        elif self.source.endswith((".tar.gz", ".tgz")):
            self.__unpack_tarfile(archive, directory, "r:gz")
        elif self.source.endswith((".tar.bz2", ".bz2")):
            self.__unpack_tarfile(archive, directory, "r:bz2")
        elif self.source.endswith((".tar.zst", ".tar.zstd")):
            self.__unpack_tarfile(archive, directory, "r:zst")
        else:
            message = f"Unknown archive type: {archive}"
            raise ValueError(message)

    @staticmethod
    def __unpack_tarfile(archive: Path, directory: Path, mode: Literal["r:gz", "r:bz2", "r:zst"]) -> None:
        if sys.version_info >= (3, 14):
            import tarfile
        else:
            # for zstd support (introduced in Python 3.14)
            # and filter kwarg (introduced in Python 3.12)
            from backports.zstd import tarfile

        with tarfile.open(archive, mode) as tf:
            tf.extractall(directory, filter="data")

    @property
    @abstractmethod
    def version(self) -> Version:
        pass

    @property
    @abstractmethod
    def python_path(self) -> str:
        pass


class CPythonStandaloneDistribution(Distribution):
    @cached_property
    def version(self) -> Version:
        from packaging.version import Version

        if (custom_version := get_custom_version(self.name)) is not None:
            return Version(f"{CUSTOM_DISTRIBUTION_VERSION_EPOCH}!{custom_version}")

        # .../cpython-3.12.0%2B20231002-...
        # .../cpython-3.7.9-...
        _, _, remaining = self.source.partition("/cpython-")
        # 3.12.0%2B20231002-...
        # 3.7.9-...
        version = remaining.split("%2B")[0] if "%2B" in remaining else remaining.split("-")[0]
        return Version(f"0!{version}")

    @cached_property
    def python_path(self) -> str:
        if (custom_path := get_custom_path(self.name)) is not None:
            return custom_path

        if self.name == "3.7":
            if sys.platform == "win32":
                return r"python\install\python.exe"

            return "python/install/bin/python3"

        if sys.platform == "win32":
            return r"python\python.exe"

        return "python/bin/python3"


class PyPyOfficialDistribution(Distribution):
    @cached_property
    def version(self) -> Version:
        from packaging.version import Version

        if (custom_version := get_custom_version(self.name)) is not None:
            return Version(f"{CUSTOM_DISTRIBUTION_VERSION_EPOCH}!{custom_version}")

        *_, remaining = self.source.partition("/pypy/")
        _, version, *_ = remaining.split("-")
        return Version(f"0!{version[1:]}")

    @cached_property
    def python_path(self) -> str:
        if (custom_path := get_custom_path(self.name)) is not None:
            return custom_path

        directory = self.archive_name
        for extension in (".tar.bz2", ".zip"):
            if directory.endswith(extension):
                directory = directory[: -len(extension)]
                break

        if sys.platform == "win32":
            return rf"{directory}\pypy.exe"

        return f"{directory}/bin/pypy"


def get_distribution(name: str, source: str = "", variant_cpu: str = "", variant_gil: str = "") -> Distribution:
    if source:
        return _get_distribution_class(source)(name, source)

    if name not in DISTRIBUTIONS:
        message = f"Unknown distribution: {name}"
        raise PythonDistributionUnknownError(message)

    arch = platform.machine().lower()
    if sys.platform == "win32":
        system = "windows"
        abi = "msvc"
    elif sys.platform == "darwin":
        system = "macos"
        abi = ""
    else:
        system = "linux"
        abi = "gnu" if any(platform.libc_ver()) else "musl"

    if not variant_cpu:
        variant_cpu = _get_default_variant_cpu(name, system, arch)

    if not variant_gil:
        variant_gil = _get_default_variant_gil()

    key = (system, arch, abi, variant_cpu, variant_gil)

    keys: dict[tuple, str] = DISTRIBUTIONS[name]
    if key not in keys:
        message = f"Could not find a default source for {name=} {system=} {arch=} {abi=} {variant_cpu=} {variant_gil=}"
        raise PythonDistributionResolutionError(message)

    source = keys[key]
    return _get_distribution_class(source)(name, source)


def get_compatible_distributions() -> dict[str, Distribution]:
    distributions: dict[str, Distribution] = {}
    for name in ORDERED_DISTRIBUTIONS:
        try:
            dist = get_distribution(name)
        except PythonDistributionResolutionError:
            pass
        else:
            distributions[name] = dist

    return distributions


def _guess_linux_variant_cpu() -> str:
    # Use the highest that we know is most common when we can't parse CPU data
    default = "v3"
    try:
        # Don't use our utility Path so we can properly mock
        with open("/proc/cpuinfo", encoding="utf-8") as f:
            contents = f.read()
    except OSError:
        return default

    # See https://clang.llvm.org/docs/UsersManual.html#x86 for the
    # instructions for each architecture variant and
    # https://github.com/torvalds/linux/blob/master/arch/x86/include/asm/cpufeatures.h
    # for the corresponding Linux flags
    v2_flags = {"cx16", "lahf_lm", "popcnt", "pni", "sse4_1", "sse4_2", "ssse3"}
    v3_flags = {"avx", "avx2", "bmi1", "bmi2", "f16c", "fma", "movbe", "xsave"} | v2_flags
    v4_flags = {"avx512f", "avx512bw", "avx512cd", "avx512dq", "avx512vl"} | v3_flags

    for line in contents.splitlines():
        key, _, value = line.partition(":")
        if key.strip() == "flags":
            flags = set(value.strip().split())

            if flags.issuperset(v4_flags):
                return "v4"

            if flags.issuperset(v3_flags):
                return "v3"

            if flags.issuperset(v2_flags):
                return "v2"

            return "v1"

    return default


def _get_default_variant_cpu(name: str, system: str, arch: str) -> str:
    # not PyPy
    if name[0].isdigit():
        variant = os.environ.get(
            "HATCH_PYTHON_VARIANT_CPU",
            # Legacy name
            os.environ.get(f"HATCH_PYTHON_VARIANT_{system.upper()}", ""),
        ).lower()

        # https://gregoryszorc.com/docs/python-build-standalone/main/running.html
        if system == "linux" and arch == "x86_64":
            # Intel-specific optimizations depending on age of release
            if variant:
                return variant

            if name == "3.8":
                return "v1"

            if name != "3.7":
                return _guess_linux_variant_cpu()

    return ""


def _get_default_variant_gil() -> str:
    return os.environ.get("HATCH_PYTHON_VARIANT_GIL", "").lower()


def _get_distribution_class(source: str) -> type[Distribution]:
    if "/python-build-standalone/releases/download/" in source:
        return CPythonStandaloneDistribution
    if source.startswith("https://downloads.python.org/pypy/"):
        return PyPyOfficialDistribution

    message = f"Unknown distribution source: {source}"
    raise ValueError(message)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Type check source code", context_settings={"ignore_unknown_options": True})
@click.argument("args", nargs=-1)
@click.option("--summarize", "-s", is_flag=True, help="Include error summary statistics")
@click.option("--cover", is_flag=True, help="Generate a type coverage report")
@click.pass_obj
def types(
    app: Application,
    *,
    args: tuple[str, ...],
    summarize: bool,
    cover: bool,
):
    """
    Type check source code, using Pyrefly by default.
    """
    from hatch.cli.types.core import TypeCheckEnvironment

    app.ensure_environment_plugin_dependencies()

    for context in app.runner_context(["hatch-check-types"]):
        tc_env = TypeCheckEnvironment(context.env)

        if cover:
            script = "coverage"
        elif summarize:
            script = "check-summarize"
        else:
            script = "check"

        default_args = tc_env.get_default_args()
        internal_args = context.env.join_command_args(default_args)
        if internal_args:
            internal_args = f" {internal_args}"

        formatted_args = context.env.join_command_args(list(args))
        context.add_shell_command(f"{script} {formatted_args}")

        context.env_vars["HATCH_CHECK_TYPES_ARGS"] = internal_args

        if not tc_env.config_path:
            tc_env.write_config_file()



import click


@click.command(short_help="Remove build artifacts")
@click.argument("location", required=False)
@click.option(
    "--target",
    "-t",
    "targets",
    multiple=True,
    help=(
        "The target with which to remove artifacts, overriding project defaults. "
        "This may be selected multiple times e.g. `-t sdist -t wheel`"
    ),
)
@click.option(
    "--hooks-only",
    is_flag=True,
    help="Whether or not to only remove artifacts from build hooks [env var: `HATCH_BUILD_HOOKS_ONLY`]",
)
@click.option(
    "--no-hooks",
    is_flag=True,
    help="Whether or not to ignore artifacts from build hooks [env var: `HATCH_BUILD_NO_HOOKS`]",
)
@click.option(
    "--ext",
    is_flag=True,
    help=(
        "Whether or not to only remove artifacts from build hooks for distributing binary Python packages, such as "
        "compiled extensions. Equivalent to `--hooks-only -t wheel`"
    ),
)
@click.pass_context
def clean(ctx, location, targets, hooks_only, no_hooks, ext):
    """Remove build artifacts."""
    from hatch.cli.build import build

    ctx.invoke(
        build, clean_only=True, location=location, targets=targets, hooks_only=hooks_only, no_hooks=no_hooks, ext=ext
    )



from __future__ import annotations

import os
import sys
from functools import cache
from importlib import import_module
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from subprocess import CompletedProcess, Popen
    from types import ModuleType

    from hatch.utils.fs import Path


@cache
def get_platform_name() -> str:
    import platform

    return normalize_platform_name(platform.system())


def normalize_platform_name(platform_name: str) -> str:
    platform_name = platform_name.lower()
    return "macos" if platform_name == "darwin" else platform_name


class Platform:
    def __init__(self, display_func: Callable = print) -> None:
        self.__display_func = display_func

        # Lazily loaded constants
        self.__default_shell: str | None = None
        self.__format_file_uri: Callable[[str], str] | None = None
        self.__join_command_args: Callable[[list[str]], str] | None = None
        self.__name: str | None = None
        self.__display_name: str | None = None
        self.__home: Path | None = None

        # Whether or not an interactive status is being displayed
        self.displaying_status = False

        self.__modules = LazilyLoadedModules()

    @property
    def modules(self) -> LazilyLoadedModules:
        """
        Accessor for lazily loading modules that either take multiple milliseconds to import
        (like `shutil` and `subprocess`) or are not used on all platforms (like `shlex`).
        """
        return self.__modules

    def format_for_subprocess(self, command: str | list[str], *, shell: bool) -> str | list[str]:
        """
        Format the given command in a cross-platform manner for immediate consumption by subprocess utilities.
        """
        if self.windows:
            # Manually locate executables on Windows to avoid multiple cases in which `shell=True` is required:
            #
            # - If the `PATH` environment variable has been modified, see:
            #   https://github.com/python/cpython/issues/52803
            # - Executables that do not have the extension `.exe`, see:
            #   https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-createprocessw
            if not shell and not isinstance(command, str):
                executable = command[0]
                new_command = [self.modules.shutil.which(executable) or executable]
                new_command.extend(command[1:])
                return new_command
        elif not shell and isinstance(command, str):
            return self.modules.shlex.split(command)

        return command

    @staticmethod
    def exit_with_code(code: str | int | None) -> None:
        sys.exit(code)

    def _run_command_integrated(
        self, command: str | list[str], *, shell: bool = False, **kwargs: Any
    ) -> CompletedProcess:
        with self.capture_process(command, shell=shell, **kwargs) as process:
            for line in self.stream_process_output(process):
                self.__display_func(line, end="")

            stdout, stderr = process.communicate()

        return self.modules.subprocess.CompletedProcess(process.args, process.poll(), stdout, stderr)

    def run_command(self, command: str | list[str], *, shell: bool = False, **kwargs: Any) -> CompletedProcess:
        """
        Equivalent to the standard library's
        [subprocess.run](https://docs.python.org/3/library/subprocess.html#subprocess.run),
        with the command first being
        [properly formatted](utilities.md#hatch.utils.platform.Platform.format_for_subprocess).
        """
        if self.displaying_status and not kwargs.get("capture_output"):
            return self._run_command_integrated(command, shell=shell, **kwargs)

        self.populate_default_popen_kwargs(kwargs, shell=shell)
        return self.modules.subprocess.run(self.format_for_subprocess(command, shell=shell), shell=shell, **kwargs)

    def check_command(self, command: str | list[str], *, shell: bool = False, **kwargs: Any) -> CompletedProcess:
        """
        Equivalent to [run_command](utilities.md#hatch.utils.platform.Platform.run_command),
        but non-zero exit codes will gracefully end program execution.
        """
        process = self.run_command(command, shell=shell, **kwargs)
        if process.returncode:
            self.exit_with_code(process.returncode)

        return process

    def check_command_output(self, command: str | list[str], *, shell: bool = False, **kwargs: Any) -> str:
        """
        Equivalent to the output from the process returned by
        [capture_process](utilities.md#hatch.utils.platform.Platform.capture_process),
        but non-zero exit codes will gracefully end program execution.
        """
        kwargs.setdefault("stdout", self.modules.subprocess.PIPE)
        kwargs.setdefault("stderr", self.modules.subprocess.STDOUT)
        self.populate_default_popen_kwargs(kwargs, shell=shell)

        process = self.modules.subprocess.run(self.format_for_subprocess(command, shell=shell), shell=shell, **kwargs)
        if process.returncode:
            # Callers might not want to merge both streams so try stderr first
            self.__display_func((process.stderr or process.stdout).decode("utf-8"))
            self.exit_with_code(process.returncode)

        return process.stdout.decode("utf-8")

    def capture_process(self, command: str | list[str], *, shell: bool = False, **kwargs: Any) -> Popen:
        """
        Equivalent to the standard library's
        [subprocess.Popen](https://docs.python.org/3/library/subprocess.html#subprocess.Popen),
        with all output captured by `stdout` and the command first being
        [properly formatted](utilities.md#hatch.utils.platform.Platform.format_for_subprocess).
        """
        self.populate_default_popen_kwargs(kwargs, shell=shell)
        return self.modules.subprocess.Popen(
            self.format_for_subprocess(command, shell=shell),
            shell=shell,
            stdout=self.modules.subprocess.PIPE,
            stderr=self.modules.subprocess.STDOUT,
            **kwargs,
        )

    def populate_default_popen_kwargs(self, kwargs: dict[str, Any], *, shell: bool) -> None:
        # https://support.apple.com/en-us/HT204899
        # https://en.wikipedia.org/wiki/System_Integrity_Protection
        if (
            "executable" not in kwargs
            and self.macos
            and shell
            and any(env_var.startswith(("DYLD_", "LD_")) for env_var in os.environ)
        ):
            default_paths = os.environ.get("PATH", os.defpath).split(os.pathsep)
            unprotected_paths = []
            for path in default_paths:
                normalized_path = os.path.normpath(path)
                if not normalized_path.startswith((
                    "/System",
                    "/usr",
                    "/bin",
                    "/sbin",
                    "/var",
                )) or normalized_path.startswith("/usr/local"):
                    unprotected_paths.append(path)

            search_path = os.pathsep.join(unprotected_paths)
            for exe_name in ("sh", "bash", "zsh", "fish"):
                executable = self.modules.shutil.which(exe_name, path=search_path)
                if executable:
                    kwargs["executable"] = executable
                    break

    @staticmethod
    def stream_process_output(process: Popen) -> Iterable[str]:
        # To avoid blocking never use a pipe's file descriptor iterator. See https://bugs.python.org/issue3907
        for line in iter(process.stdout.readline, b""):  # type: ignore[union-attr]
            yield line.decode("utf-8")

    @property
    def default_shell(self) -> str:
        """
        Returns the default shell of the system.

        On Windows systems first try the `SHELL` environment variable, if present, followed by
        the `COMSPEC` environment variable, defaulting to `cmd`. On all other platforms only
        the `SHELL` environment variable will be used, defaulting to `bash`.
        """
        if self.__default_shell is None:
            if self.windows:
                self.__default_shell = cast(str, os.environ.get("SHELL", os.environ.get("COMSPEC", "cmd")))
            else:
                self.__default_shell = cast(str, os.environ.get("SHELL", "bash"))
        return self.__default_shell

    @property
    def join_command_args(self) -> Callable[[list[str]], str]:
        if self.__join_command_args is None:
            if self.windows:
                self.__join_command_args = self.modules.subprocess.list2cmdline
            else:
                self.__join_command_args = self.modules.shlex.join

        return self.__join_command_args

    @property
    def format_file_uri(self) -> Callable[[str], str]:
        if self.__format_file_uri is None:
            if self.windows:
                self.__format_file_uri = lambda p: f"file:///{p}".replace("\\", "/")
            else:
                self.__format_file_uri = lambda p: f"file://{p}"

        return self.__format_file_uri

    @property
    def windows(self) -> bool:
        """
        Indicates whether Hatch is running on Windows.
        """
        return self.name == "windows"

    @property
    def macos(self) -> bool:
        """
        Indicates whether Hatch is running on macOS.
        """
        return self.name == "macos"

    @property
    def linux(self) -> bool:
        """
        Indicates whether Hatch is running on neither Windows nor macOS.
        """
        return not (self.windows or self.macos)

    def exit_with_command(self, command: list[str]) -> None:
        """
        Run the given command and exit with its exit code. On non-Windows systems, this uses the standard library's
        [os.execvp](https://docs.python.org/3/library/os.html#os.execvp).
        """
        if self.windows:
            process = self.run_command(command)
            self.exit_with_code(process.returncode)
        else:
            os.execvp(command[0], command)  # noqa: S606

    @property
    def name(self) -> str:
        """
        One of the following:

        - `linux`
        - `windows`
        - `macos`
        """
        if self.__name is None:
            self.__name = get_platform_name()

        return self.__name

    @property
    def display_name(self) -> str:
        """
        One of the following:

        - `Linux`
        - `Windows`
        - `macOS`
        """
        if self.__display_name is None:
            self.__display_name = "macOS" if self.macos else self.name.capitalize()

        return self.__display_name

    @property
    def home(self) -> Path:
        """
        The user's home directory as a path-like object.
        """
        if self.__home is None:
            from hatch.utils.fs import Path

            self.__home = Path(os.path.expanduser("~"))

        return self.__home


class LazilyLoadedModules:
    def __getattr__(self, name: str) -> ModuleType:
        module = import_module(name)
        setattr(self, name, module)
        return module



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(
    short_help="Restore the installation", context_settings={"help_option_names": [], "ignore_unknown_options": True}
)
@click.argument("args", nargs=-1)
@click.pass_obj
def restore(app: Application, *, args: tuple[str, ...]):  # noqa: ARG001
    app.abort("Hatch is not installed as a binary")



from __future__ import annotations

import os
from typing import cast

import click

from hatch._version import __version__
from hatch.cli.application import Application
from hatch.cli.build import build
from hatch.cli.check import check
from hatch.cli.clean import clean
from hatch.cli.config import config
from hatch.cli.dep import dep
from hatch.cli.env import env
from hatch.cli.fmt import fmt
from hatch.cli.lock_cmd import lock_command
from hatch.cli.new import new
from hatch.cli.project import project
from hatch.cli.publish import publish
from hatch.cli.python import python
from hatch.cli.run import run
from hatch.cli.self import self_command
from hatch.cli.shell import shell
from hatch.cli.status import status
from hatch.cli.test import test
from hatch.cli.version import version
from hatch.config.constants import AppEnvVars, ConfigEnvVars
from hatch.project.core import Project
from hatch.utils.ci import running_in_ci
from hatch.utils.fs import Path


@click.group(
    context_settings={"help_option_names": ["-h", "--help"], "max_content_width": 120}, invoke_without_command=True
)
@click.option(
    "--env",
    "-e",
    "env_name",
    envvar=AppEnvVars.ENV,
    default="default",
    help="The name of the environment to use [env var: `HATCH_ENV`]",
)
@click.option(
    "--project",
    "-p",
    envvar=ConfigEnvVars.PROJECT,
    help="The name of the project to work on [env var: `HATCH_PROJECT`]",
)
@click.option(
    "--verbose",
    "-v",
    envvar=AppEnvVars.VERBOSE,
    count=True,
    help="Increase verbosity (can be used additively) [env var: `HATCH_VERBOSE`]",
)
@click.option(
    "--quiet",
    "-q",
    envvar=AppEnvVars.QUIET,
    count=True,
    help="Decrease verbosity (can be used additively) [env var: `HATCH_QUIET`]",
)
@click.option(
    "--color/--no-color",
    default=None,
    help="Whether or not to display colored output (default is auto-detection) [env vars: `FORCE_COLOR`/`NO_COLOR`]",
)
@click.option(
    "--interactive/--no-interactive",
    envvar=AppEnvVars.INTERACTIVE,
    default=None,
    help=(
        "Whether or not to allow features like prompts and progress bars (default is auto-detection) "
        "[env var: `HATCH_INTERACTIVE`]"
    ),
)
@click.option(
    "--data-dir",
    envvar=ConfigEnvVars.DATA,
    help="The path to a custom directory used to persist data [env var: `HATCH_DATA_DIR`]",
)
@click.option(
    "--cache-dir",
    envvar=ConfigEnvVars.CACHE,
    help="The path to a custom directory used to cache data [env var: `HATCH_CACHE_DIR`]",
)
@click.option(
    "--config",
    "config_file",
    envvar=ConfigEnvVars.CONFIG,
    help="The path to a custom config file to use [env var: `HATCH_CONFIG`]",
)
@click.version_option(version=__version__, prog_name="Hatch")
@click.pass_context
def hatch(
    ctx: click.Context,
    env_name,
    project,
    verbose,
    quiet,
    color,
    interactive,
    data_dir,
    cache_dir,
    config_file,
):
    """
    \b
     _   _       _       _
    | | | |     | |     | |
    | |_| | __ _| |_ ___| |__
    |  _  |/ _` | __/ __| '_ \\
    | | | | (_| | || (__| | | |
    \\_| |_/\\__,_|\\__\\___|_| |_|
    """
    if color is None:
        if os.environ.get(AppEnvVars.NO_COLOR) == "1":
            color = False
        elif os.environ.get(AppEnvVars.FORCE_COLOR) == "1":
            color = True

    if interactive is None and running_in_ci():
        interactive = False

    app = Application(ctx.exit, verbosity=verbose - quiet, enable_color=color, interactive=interactive)
    app.env_active = os.environ.get(AppEnvVars.ENV_ACTIVE)
    if (
        app.env_active
        and (param_source := ctx.get_parameter_source("env_name")) is not None
        and param_source.name == "DEFAULT"
    ):
        app.env = app.env_active
    else:
        app.env = env_name

    if config_file:
        app.config_file.path = Path(config_file).resolve()
        if not app.config_file.path.is_file():
            app.abort(f"The selected config file `{app.config_file.path}` does not exist.")
    elif not app.config_file.path.is_file():
        if app.verbose:
            app.display_waiting("No config file found, creating one with default settings now...")

        try:
            app.config_file.restore()
            if app.verbose:
                app.display_success("Success! Please see `hatch config`.")
        except OSError:  # no cov
            app.abort(
                f"Unable to create config file located at `{app.config_file.path}`. Please check your permissions."
            )

    if not ctx.invoked_subcommand:
        app.display_info(ctx.get_help())
        return

    # Persist app data for sub-commands
    ctx.obj = app

    try:
        app.config_file.load()
    except OSError as e:  # no cov
        app.abort(f"Error loading configuration: {e}")

    app.config.terminal.styles.parse_fields()
    errors = app.initialize_styles(app.config.terminal.styles.raw_data)
    if errors and color is not False and not app.quiet:  # no cov
        for error in errors:
            app.display_warning(error)

    app.data_dir = Path(data_dir or app.config.dirs.data).expand()
    app.cache_dir = Path(cache_dir or app.config.dirs.cache).expand()

    if project:
        potential_project = Project.from_config(app.config, project)
        if potential_project is None or potential_project.root is None:
            app.abort(f"Unable to locate project {project}")

        app.project = cast(Project, potential_project)
        app.project.set_app(app)
        return

    app.project = Project(Path.cwd())
    app.project.set_app(app)

    if app.config.mode == "local":
        return

    # The following logic is mostly duplicated for each branch so coverage can be asserted
    if app.config.mode == "project":
        if not app.config.project:
            app.display_warning("Mode is set to `project` but no project is set, defaulting to the current directory")
            return

        possible_project = Project.from_config(app.config, app.config.project)
        if possible_project is None:
            app.display_warning(f"Unable to locate project {app.config.project}, defaulting to the current directory")
        else:
            app.project = possible_project

        app.project.set_app(app)
        return

    if app.config.mode == "aware" and app.project.root is None:
        if not app.config.project:
            app.display_warning("Mode is set to `aware` but no project is set, defaulting to the current directory")
            return

        possible_project = Project.from_config(app.config, app.config.project)
        if possible_project is None:
            app.display_warning(f"Unable to locate project {app.config.project}, defaulting to the current directory")
        else:
            app.project = possible_project

        app.project.set_app(app)
        return


hatch.add_command(build)
hatch.add_command(check)
hatch.add_command(clean)
hatch.add_command(config)
hatch.add_command(dep)
hatch.add_command(env)
hatch.add_command(fmt)
hatch.add_command(lock_command)
hatch.add_command(new)
hatch.add_command(project)
hatch.add_command(publish)
hatch.add_command(python)
hatch.add_command(run)
hatch.add_command(self_command)
hatch.add_command(shell)
hatch.add_command(status)
hatch.add_command(test)
hatch.add_command(version)


def main():  # no cov
    try:
        hatch(prog_name="hatch", windows_expand_args=False)
    except Exception:  # noqa: BLE001
        import sys

        from rich.console import Console

        console = Console()
        hatch_debug = os.getenv("HATCH_DEBUG") in {"1", "true"}
        console.print_exception(suppress=[click], show_locals=hatch_debug)
        sys.exit(1)



from abc import ABC, abstractmethod

from hatch.env.utils import get_verbosity_flag
from hatchling.utils.context import ContextFormatter


class EnvironmentContextFormatterBase(ContextFormatter, ABC):
    @abstractmethod
    def formatters(self):
        return {}


class EnvironmentContextFormatter(EnvironmentContextFormatterBase):
    def __init__(self, environment):
        self.environment = environment
        self.CONTEXT_NAME = f"environment_{environment.PLUGIN_NAME}"

    def formatters(self):  # noqa: PLR6301
        """
        This returns a mapping of supported field names to their respective formatting functions. Each function
        accepts 2 arguments:

        - the `value` that was passed to the format call, defaulting to `None`
        - the modifier `data`, defaulting to an empty string
        """
        return {}

    def get_formatters(self):
        formatters = {
            "args": self.__format_args,
            "env_name": self.__format_env_name,
            "env_type": self.__format_env_type,
            "matrix": self.__format_matrix,
            "verbosity": self.__format_verbosity,
        }
        formatters.update(self.formatters())
        return formatters

    def __format_args(self, value, data):  # noqa: PLR6301
        if value is not None:
            return value

        return data or ""

    def __format_env_name(self, value, data):  # noqa: ARG002
        return self.environment.name

    def __format_env_type(self, value, data):  # noqa: ARG002
        return self.environment.PLUGIN_NAME

    def __format_matrix(self, value, data):  # noqa: ARG002
        if not data:
            message = "The `matrix` context formatting field requires a modifier"
            raise ValueError(message)

        variable, separator, default = data.partition(":")
        if variable in self.environment.matrix_variables:
            return self.environment.matrix_variables[variable]

        if not separator:
            message = f"Nonexistent matrix variable must set a default: {variable}"
            raise ValueError(message)

        return default

    def __format_verbosity(self, value, data):  # noqa: ARG002
        if not data:
            return str(self.environment.verbosity)

        modifier, _, adjustment = data.partition(":")
        if modifier != "flag":
            message = f"Unknown verbosity modifier: {modifier}"
            raise ValueError(message)

        if not adjustment:
            adjustment = "0"

        try:
            adjustment = int(adjustment)
        except ValueError:
            message = f"Verbosity flag adjustment must be an integer: {adjustment}"
            raise TypeError(message) from None

        return get_verbosity_flag(self.environment.verbosity, adjustment=adjustment)



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from hatch.utils.fs import Path


class File:
    def __init__(self, path: Path | None, contents: str = ""):
        self.path = path
        self.contents = contents
        self.feature = None

    def write(self, root):
        if self.path is None:  # no cov
            return

        path = root / self.path
        path.ensure_parent_dir_exists()
        path.write_text(self.contents, encoding="utf-8")


def find_template_files(module):
    for name in dir(module):
        obj = getattr(module, name)
        if obj is File:
            continue

        with suppress(TypeError):
            if issubclass(obj, File):
                yield obj



from __future__ import annotations

from typing import TYPE_CHECKING

from hatch.env.collectors.custom import CustomEnvironmentCollector
from hatch.env.collectors.default import DefaultEnvironmentCollector
from hatchling.plugin import hookimpl

if TYPE_CHECKING:
    from hatch.env.collectors.plugin.interface import EnvironmentCollectorInterface


@hookimpl
def hatch_register_environment_collector() -> list[type[EnvironmentCollectorInterface]]:
    return [CustomEnvironmentCollector, DefaultEnvironmentCollector]



from __future__ import annotations

from functools import cached_property

from packaging.requirements import InvalidRequirement, Requirement

from hatch.utils.fs import Path

InvalidDependencyError = InvalidRequirement


class Dependency(Requirement):
    def __init__(self, s: str, *, editable: bool = False) -> None:
        super().__init__(s)

        if editable and self.url is None:
            message = f"Editable dependency must refer to a local path: {s}"
            raise InvalidDependencyError(message)

        self.__editable = editable

    @property
    def editable(self) -> bool:
        return self.__editable

    @cached_property
    def path(self) -> Path | None:
        from urllib.parse import unquote

        if self.url is None:
            return None

        import hyperlink

        uri = hyperlink.parse(self.url)
        if uri.scheme != "file":
            return None
        decoded_url = unquote(self.url)
        return Path.from_uri(decoded_url)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="View or set a project's version")
@click.argument("desired_version", required=False)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Allow an explicit downgrading version to be given",
)
@click.pass_obj
def version(app: Application, *, desired_version: str | None, force: bool):
    """View or set a project's version."""
    if app.project.root is None:
        if app.project.chosen_name is None:
            app.abort("No project detected")
        else:
            app.abort(f"Project {app.project.chosen_name} (not a project)")

    if "version" in app.project.metadata.config.get("project", {}):
        if desired_version:
            app.abort("Cannot set version when it is statically defined by the `project.version` field")
        else:
            app.display(app.project.metadata.config["project"]["version"])
            return

    from hatch.config.constants import VersionEnvVars
    from hatch.project.constants import BUILD_BACKEND

    with app.project.location.as_cwd():
        if app.project.metadata.build.build_backend != BUILD_BACKEND:
            if desired_version:
                app.abort("The version can only be set when Hatchling is the build backend")

            app.ensure_environment_plugin_dependencies()
            app.project.prepare_build_environment()

            with app.project.location.as_cwd(), app.project.build_env.get_env_vars():
                project_metadata = app.project.build_frontend.get_core_metadata()

            app.display(project_metadata["version"])
        else:
            from hatch.utils.runner import ExecutionContext

            app.ensure_environment_plugin_dependencies()
            app.project.prepare_build_environment()

            context = ExecutionContext(app.project.build_env)
            command = ["python", "-u", "-m", "hatchling", "version"]
            if desired_version:
                command.append(desired_version)
                if force:
                    context.env_vars[VersionEnvVars.VALIDATE_BUMP] = "false"

            context.add_shell_command(command)
            app.execute_context(context)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Remove environments")
@click.argument("env_name", default="default")
@click.pass_context
def remove(ctx: click.Context, env_name: str):
    """Remove environments."""
    app: Application = ctx.obj
    app.ensure_environment_plugin_dependencies()

    if (parameter_source := ctx.get_parameter_source("env_name")) is not None and parameter_source.name == "DEFAULT":
        env_name = app.env

    environments = app.project.expand_environments(env_name)
    if not environments:
        app.abort(f"Environment `{env_name}` is not defined by project config")

    for env_name in environments:
        if env_name == app.env_active:
            app.abort(f"Cannot remove active environment: {env_name}")

    for env_name in environments:
        environment = app.project.get_environment(env_name)
        if environment.exists():
            with app.status(f"Removing environment: {env_name}"):
                environment.remove()



from hatch.template import File, files_default, find_template_files
from hatch.template.plugin.interface import TemplateInterface
from hatch.utils.fs import Path
from hatch.utils.network import download_file


class DefaultTemplate(TemplateInterface):
    PLUGIN_NAME = "default"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.plugin_config.setdefault("ci", False)
        self.plugin_config.setdefault("src-layout", True)
        self.plugin_config.setdefault("tests", True)

    def initialize_config(self, config):
        # Default values
        config["readme_file_path"] = "README.md"
        config["package_metadata_file_path"] = f"src/{config['package_name']}/__about__.py"

        license_data = {}

        # Licenses
        license_ids = config["licenses"]["default"]
        if not license_ids:
            config["license_data"] = license_data
            config["license_expression"] = ""
            config["license_files"] = ""
            config["license_header"] = ""
            return

        cached_licenses_dir = self.cache_dir / "licenses"
        cached_licenses_dir.ensure_dir_exists()

        license_ids = sorted(set(license_ids))
        for license_id in sorted(set(license_ids)):
            license_file_name = f"{license_id}.txt"
            cached_license_path = cached_licenses_dir / license_file_name
            if not cached_license_path.is_file():
                from packaging.licenses._spdx import VERSION  # noqa: PLC2701

                url = f"https://raw.githubusercontent.com/spdx/license-list-data/v{VERSION}/text/{license_file_name}"
                for _ in range(5):
                    try:
                        download_file(cached_license_path, url)
                    except Exception:  # noqa: BLE001, S112
                        continue
                    else:
                        break

            license_data[license_id] = cached_license_path.read_text(encoding="utf-8")

        config["license_data"] = license_data
        config["license_expression"] = " OR ".join(license_data)
        config["license_header"] = (
            ""
            if not config["licenses"]["headers"]
            else f"""\
# SPDX-FileCopyrightText: {self.creation_time.year}-present {config["name"]} <{config["email"]}>
#
# SPDX-License-Identifier: {config["license_expression"]}
"""
        )
        if len(license_ids) == 1:
            config["license_files"] = ""
        else:
            config["license_files"] = '\nlicense-files = { globs = ["LICENSES/*"] }'

        if config["args"]["cli"]:
            config["dependencies"].add("click")

        if not self.plugin_config["src-layout"]:
            config["package_metadata_file_path"] = f"{config['package_metadata_file_path'][4:]}"

    def get_files(self, config):
        files = list(find_template_files(files_default))

        # Add any licenses
        license_data = config["license_data"]
        if license_data:
            if len(license_data) == 1:
                license_id, text = next(iter(license_data.items()))
                license_text = get_license_text(config, license_id, text, self.creation_time)
                files.append(File(Path("LICENSE.txt"), license_text))
            else:
                # https://reuse.software/faq/#multi-licensing
                for license_id, text in license_data.items():
                    license_text = get_license_text(config, license_id, text, self.creation_time)
                    files.append(File(Path("LICENSES", f"{license_id}.txt"), license_text))

        if config["args"]["cli"]:
            from hatch.template import files_feature_cli

            files.extend(find_template_files(files_feature_cli))

        if self.plugin_config["tests"]:
            from hatch.template import files_feature_tests

            files.extend(find_template_files(files_feature_tests))

        if self.plugin_config["ci"]:
            from hatch.template import files_feature_ci

            files.extend(find_template_files(files_feature_ci))

        return files

    def finalize_files(self, config, files):
        if config["licenses"]["headers"] and config["license_data"]:
            for template_file in files:
                if template_file.path.name.endswith(".py"):
                    template_file.contents = config["license_header"] + template_file.contents

        if self.plugin_config["src-layout"]:
            for template_file in files:
                if template_file.path.parts[0] == config["package_name"]:
                    template_file.path = Path("src", template_file.path)


def get_license_text(config, license_id, license_text, creation_time):
    if license_id == "MIT":
        license_text = license_text.replace("<year>", f"{creation_time.year}-present", 1)
        license_text = license_text.replace("<copyright holders>", f"{config['name']} <{config['email']}>", 1)
    elif license_id == "BSD-3-Clause":
        license_text = license_text.replace("<year>", f"{creation_time.year}-present", 1)
        license_text = license_text.replace("<owner>", f"{config['name']} <{config['email']}>", 1)

    return f"{license_text.rstrip()}\n"



from __future__ import annotations

from hatch.utils.fs import Path


class AuthenticationCredentials:
    def __init__(
        self,
        app,
        cache_dir: Path,
        options: dict,
        repo: str,
        repo_config: dict[str, str],
    ):
        self._app = app
        self._pwu_path = cache_dir / "previous_working_users.json"
        self._options = options
        self._repo = repo
        self._repo_config = repo_config

        self.__username: str | None = None
        self.__password: str | None = None
        self.__username_was_read = False
        self.__password_was_read = False
        self.__pwu_data: dict[str, str] = {}

    @property
    def password(self) -> str:
        if self.__password is None:
            self.__password = self.__get_password()
        return self.__password

    @property
    def username(self) -> str:
        if self.__username is None:
            self.__username = self.__get_username()
        return self.__username

    def __get_password(self) -> str:
        # this method doesn't consider .pypirc as the __password attribute would have
        # been set when it was looked up during username retrieval

        password = self._options.get("auth") or self._repo_config.get("auth")
        if password is not None:
            return password

        import keyring

        keyring_service = self._repo_config["url"]
        password = keyring.get_password(keyring_service, self.username)
        if password is not None:
            return password

        if self._options["no_prompt"]:
            self._app.abort("Missing required option: auth")

        self.__password_was_read = True
        return self._app.prompt("Password / Token", hide_input=True)

    def __get_username(self) -> str:
        username = (
            self._options.get("user")
            or self._repo_config.get("user")
            or self._read_pypirc()
            or self._read_previous_working_user_data()
        )
        if username is not None:
            return username

        if self._options["no_prompt"]:
            self._app.abort("Missing required option: user")

        self.__username_was_read = True
        return self._app.prompt(f"Username for '{self._repo_config['url']}' [__token__]") or "__token__"

    def _read_previous_working_user_data(self) -> str | None:
        if self._pwu_path.is_file():
            contents = self._pwu_path.read_text()
            if contents:
                import json

                self.__pwu_data = json.loads(contents)
        return self.__pwu_data.get(self._repo)

    def _read_pypirc(self) -> str | None:
        import configparser

        pypirc = configparser.ConfigParser()
        pypirc.read(Path.home() / ".pypirc")
        repo = self._repo or "pypi"

        if pypirc.has_section(repo):
            self.__password = pypirc.get(section=repo, option="password", fallback=None)
            return pypirc.get(section=repo, option="username", fallback=None)

        repo_url = self._repo_config["url"]
        for section in pypirc.sections():
            if pypirc.get(section=section, option="repository", fallback=None) == repo_url:
                self.__password = pypirc.get(section=section, option="password", fallback=None)
                return pypirc.get(section=section, option="username", fallback=None)

        return None

    def write_updated_data(self):
        if self.__username_was_read:
            import json

            self.__pwu_data[self._repo] = self.__username
            self._pwu_path.ensure_parent_dir_exists()
            self._pwu_path.write_text(json.dumps(self.__pwu_data))

        if self.__password_was_read:
            import keyring

            keyring_service = self._repo_config["url"]
            keyring.set_password(keyring_service, self.__username, self.__password)



from hatch.template import File
from hatch.utils.fs import Path


class CommandLinePackage(File):
    TEMPLATE = """\
name: test

on:
  push:
    branches: [main, master]
  pull_request:
    branches: [main, master]

concurrency:
  group: test-${{ github.head_ref }}
  cancel-in-progress: true

env:
  PYTHONUNBUFFERED: "1"
  FORCE_COLOR: "1"

jobs:
  run:
    name: Python ${{ matrix.python-version }} on ${{ startsWith(matrix.os, 'macos-') && 'macOS' || startsWith(matrix.os, 'windows-') && 'Windows' || 'Linux' }}
    runs-on: ${{ matrix.os }}
    strategy:
      fail-fast: false
      matrix:
        os: [ubuntu-latest, windows-latest, macos-latest]
        python-version: ['3.8', '3.9', '3.10', '3.11', '3.12', '3.13']

    steps:
    - uses: actions/checkout@v3

    - name: Set up Python ${{ matrix.python-version }}
      uses: actions/setup-python@v4
      with:
        python-version: ${{ matrix.python-version }}

    - name: Install Hatch
      run: pip install --upgrade hatch

    - name: Run static analysis
      run: hatch fmt --check

    - name: Run tests
      run: hatch test --python ${{ matrix.python-version }} --cover --randomize --parallel --retries 2 --retry-delay 1
"""

    def __init__(
        self,
        template_config: dict,  # noqa: ARG002
        plugin_config: dict,  # noqa: ARG002
    ):
        super().__init__(Path(".github", "workflows", "test.yml"), self.TEMPLATE)



from __future__ import annotations

import os
from functools import cached_property
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from hatch.env.plugin.interface import EnvironmentInterface
    from hatch.utils.fs import Path


class TypeCheckEnvironment:
    def __init__(self, env: EnvironmentInterface) -> None:
        self.env = env

    @cached_property
    def config_path(self) -> str:
        return self.env.config.get("config-path", "")

    def get_default_args(self) -> list[str]:
        default_args: list[str] = []
        if not self.config_path and not self.user_config_file:
            default_args.extend(["--config", str(self.internal_config_file)])
        return default_args

    @cached_property
    def internal_config_file(self) -> Path:
        return self.env.isolated_data_directory / ".config" / self.env.root.id / "pyrefly.toml"

    @cached_property
    def user_config_file(self) -> Path | None:
        """Check if the user already has a pyrefly config file."""
        # Check for standalone pyrefly.toml
        if (config_file := (self.env.root / "pyrefly.toml")).is_file():
            return config_file

        # Check for [tool.pyrefly] in pyproject.toml
        pyproject = self.env.root / "pyproject.toml"
        if pyproject.is_file():
            from hatch.utils.toml import load_toml_data

            data = load_toml_data(pyproject.read_text())
            if "tool" in data and "pyrefly" in data["tool"]:
                return pyproject

        return None

    def construct_config_file(self) -> str:
        """Generate a minimal pyrefly.toml based on the project layout.

        The generated config tells pyrefly:
        - Where to find source code (project-includes)
        - Where to resolve imports from (search-path)
        - To use relaxed type checking rules (preset = "legacy")
        - To not guess import roots on its own (disable-search-path-heuristics)
        - Which platform-specific modules to ignore (ignore-missing-imports)
        """
        lines: list[str] = []
        root = str(self.env.root)

        project_includes = self._detect_project_includes()
        search_paths = self._detect_search_paths()
        ignore_imports = self._detect_platform_conditional_deps()

        # Paths must be absolute because the auto-generated config file is stored in an
        # isolated data directory (not the project root), so relative paths would not resolve
        # to the correct project directories.
        if project_includes:
            abs_includes = [os.path.join(root, p) for p in project_includes]
            includes_str = ", ".join(f'"{p}"' for p in abs_includes)
            # project-includes tells pyrefly which directories contain source files to check
            lines.append(f"project-includes = [{includes_str}]")

        if search_paths:
            abs_paths = [os.path.join(root, p) for p in search_paths]
            paths_str = ", ".join(f'"{p}"' for p in abs_paths)
            # search-path tells pyrefly where to resolve imports from (e.g., "src" for src-layout)
            lines.append(f"search-path = [{paths_str}]")

        # "legacy" preset uses relaxed type checking rules suitable for existing codebases
        # that haven't been fully annotated yet.
        # Disable pyrefly's automatic search path detection since we explicitly configure paths
        # above; without this, pyrefly may add duplicate or incorrect import roots.
        lines.extend(('preset = "legacy"', "disable-search-path-heuristics = true"))

        if ignore_imports:
            imports_str = ", ".join(f'"{m}"' for m in ignore_imports)
            # Suppress errors for platform-conditional dependencies that aren't installed
            # on the current OS (e.g., a Linux-only package when running on macOS)
            lines.append(f"ignore-missing-imports = [{imports_str}]")

        # Ensure file ends with newline
        lines.append("")
        return "\n".join(lines)

    def write_config_file(self) -> None:
        """Write the auto-generated config file."""
        if self.user_config_file:
            return

        config_contents = self.construct_config_file()

        if self.config_path:
            (self.env.root / self.config_path).write_atomic(config_contents, "w", encoding="utf-8")
            return

        self.internal_config_file.parent.ensure_dir_exists()
        self.internal_config_file.write_text(config_contents)

    def _detect_project_includes(self) -> list[str]:
        """Detect which directories contain source code to type check.

        This is used to populate pyrefly's `project-includes` config option, which tells
        pyrefly which directories to scan for Python files. Without this, pyrefly would
        either check nothing (if project-includes is empty) or fall back to its own
        heuristics which may miss source in non-standard layouts like src-layout or
        monorepo workspace members.
        """
        includes: list[str] = []
        root = str(self.env.root)

        # Standard src layout for the main project
        src_dir = os.path.join(root, "src")
        if os.path.isdir(src_dir):
            for entry in os.listdir(src_dir):
                entry_path = os.path.join(src_dir, entry)
                if os.path.isdir(entry_path) and os.path.isfile(os.path.join(entry_path, "__init__.py")):
                    includes.append(f"src/{entry}")

        # Workspace members (monorepo support)
        for member_path in self._get_workspace_member_paths():
            relative = os.path.relpath(member_path, root)
            member_src = os.path.join(member_path, "src")
            if os.path.isdir(member_src):
                for entry in os.listdir(member_src):
                    entry_path = os.path.join(member_src, entry)
                    if os.path.isdir(entry_path) and os.path.isfile(os.path.join(entry_path, "__init__.py")):
                        includes.append(f"{relative}/src/{entry}")

        # Tests directory
        tests_dir = os.path.join(root, "tests")
        if os.path.isdir(tests_dir) and os.path.isfile(os.path.join(tests_dir, "__init__.py")):
            includes.append("tests")

        return includes

    def _detect_search_paths(self) -> list[str]:
        """Detect import root directories for pyrefly's search-path."""
        paths: list[str] = []
        root = str(self.env.root)

        # src/ layout means imports resolve from src/
        if os.path.isdir(os.path.join(root, "src")):
            paths.append("src")

        # Workspace members
        for member_path in self._get_workspace_member_paths():
            relative = os.path.relpath(member_path, root)
            member_src = os.path.join(member_path, "src")
            if os.path.isdir(member_src):
                paths.append(f"{relative}/src")

        # tests/ as a package root (for relative imports within tests)
        tests_dir = os.path.join(root, "tests")
        if os.path.isdir(tests_dir) and os.path.isfile(os.path.join(tests_dir, "__init__.py")):
            paths.append("tests")

        return paths

    def _get_workspace_member_paths(self) -> list[str]:
        """Get filesystem paths of workspace members from the environment config."""
        try:
            members = self.env.workspace.members
        except Exception:  # noqa: BLE001
            return []

        return [str(member.project.location) for member in members]

    def _detect_platform_conditional_deps(self) -> list[str]:
        """Detect dependencies with platform markers that won't be installed locally."""
        import sys

        ignore: list[str] = []

        try:
            project_config = self.env.metadata.config
        except Exception:  # noqa: BLE001
            return ignore

        dependencies = project_config.get("project", {}).get("dependencies", [])
        if not isinstance(dependencies, list):
            return ignore

        for dep in dependencies:
            if not isinstance(dep, str):
                continue

            # Simple heuristic: if a dependency has a platform marker that doesn't match
            # the current platform, add its import name to ignore list
            if "sys_platform" in dep:
                dep_lower = dep.lower()
                current_platform = sys.platform

                # Check if the marker excludes the current platform
                if (
                    ('sys_platform == "linux"' in dep_lower and current_platform != "linux")
                    or ('sys_platform == "win32"' in dep_lower and current_platform != "win32")
                    or ('sys_platform == "darwin"' in dep_lower and current_platform != "darwin")
                ):
                    ignore.append(self._dep_to_import_name(dep))

        return ignore

    @staticmethod
    def _dep_to_import_name(dep: str) -> str:
        """Convert a dependency string to its likely import name."""
        # Strip markers (everything after ;)
        name = dep.split(";")[0].strip()
        # Strip version specifiers
        for char in (">=", "<=", "!=", "~=", "==", ">", "<", "["):
            name = name.split(char)[0]
        # Normalize: dashes become underscores in import names
        return name.strip().replace("-", "_").lower()

    @cached_property
    def user_config(self) -> dict[str, Any]:
        if self.user_config_file is None:
            return {}

        from hatch.utils.toml import load_toml_data

        data = load_toml_data(self.user_config_file.read_text())

        if self.user_config_file.name == "pyproject.toml":
            return data.get("tool", {}).get("pyrefly", {})

        return data



from __future__ import annotations

import os
import sys
from abc import ABC, abstractmethod
from contextlib import contextmanager
from functools import cached_property
from os.path import isabs
from typing import TYPE_CHECKING, Any

from hatch.config.constants import AppEnvVars
from hatch.env.utils import add_verbosity_flag, get_env_var_option
from hatch.project.utils import format_script_commands, parse_script_command
from hatch.utils.structures import EnvVars

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    from hatch.dep.core import Dependency
    from hatch.project.core import Project
    from hatch.utils.fs import Path


class EnvironmentInterface(ABC):
    """
    Example usage:

    ```python tab="plugin.py"
    from hatch.env.plugin.interface import EnvironmentInterface


    class SpecialEnvironment(EnvironmentInterface):
        PLUGIN_NAME = "special"
        ...
    ```

    ```python tab="hooks.py"
    from hatchling.plugin import hookimpl

    from .plugin import SpecialEnvironment


    @hookimpl
    def hatch_register_environment():
        return SpecialEnvironment
    ```
    """

    PLUGIN_NAME = ""
    """The name used for selection."""

    def __init__(
        self,
        root,
        metadata,
        name,
        config,
        matrix_variables,
        data_directory,
        isolated_data_directory,
        platform,
        verbosity,
        app,
    ):
        self.__root = root
        self.__metadata = metadata
        self.__name = name
        self.__config = config
        self.__matrix_variables = matrix_variables
        self.__data_directory = data_directory
        self.__isolated_data_directory = isolated_data_directory
        self.__platform = platform
        self.__verbosity = verbosity
        self.__app = app

        self.additional_dependencies = []

    @property
    def matrix_variables(self):
        return self.__matrix_variables

    @property
    def app(self):
        """
        An instance of [Application](../utilities.md#hatchling.bridge.app.Application).
        """
        return self.__app

    @cached_property
    def context(self):
        return self.get_context()

    @property
    def verbosity(self):
        return self.__verbosity

    @property
    def root(self):
        """
        The root of the local project tree as a path-like object.
        """
        return self.__root

    @property
    def metadata(self):
        return self.__metadata

    @property
    def name(self) -> str:
        """
        The name of the environment.
        """
        return self.__name

    @property
    def platform(self):
        """
        An instance of [Platform](../utilities.md#hatch.utils.platform.Platform).
        """
        return self.__platform

    @property
    def data_directory(self):
        """
        The [directory](../../config/hatch.md#environments) this plugin should use for storage as a path-like object.
        If the user has not configured one then this will be the same as the
        [isolated data directory](reference.md#hatch.env.plugin.interface.EnvironmentInterface.isolated_data_directory).
        """
        return self.__data_directory

    @property
    def isolated_data_directory(self):
        """
        The default [directory](../../config/hatch.md#environments) reserved exclusively for this plugin as a path-like
        object.
        """
        return self.__isolated_data_directory

    @property
    def config(self) -> dict:
        """
        ```toml config-example
        [tool.hatch.envs.<ENV_NAME>]
        ```
        """
        return self.__config

    @cached_property
    def project_root(self) -> str:
        """
        The root of the project tree as a string. If the environment is not running locally,
        this should be the remote path to the project.
        """
        return str(self.root)

    @cached_property
    def sep(self) -> str:
        """
        The character used to separate directories in paths. By default, this is `\\` on Windows and `/` otherwise.
        """
        return os.sep

    @cached_property
    def pathsep(self) -> str:
        """
        The character used to separate paths. By default, this is `;` on Windows and `:` otherwise.
        """
        return os.pathsep

    @cached_property
    def system_python(self) -> str:
        system_python = os.environ.get(AppEnvVars.PYTHON)
        if system_python == "self":
            system_python = sys.executable

        system_python = (
            system_python
            or self.platform.modules.shutil.which("python")
            or self.platform.modules.shutil.which("python3")
            or sys.executable
        )
        if not isabs(system_python):
            system_python = self.platform.modules.shutil.which(system_python)

        return system_python

    @cached_property
    def env_vars(self) -> dict[str, str]:
        """
        ```toml config-example
        [tool.hatch.envs.<ENV_NAME>.env-vars]
        ```
        """
        env_vars = self.config.get("env-vars", {})
        if not isinstance(env_vars, dict):
            message = f"Field `tool.hatch.envs.{self.name}.env-vars` must be a mapping"
            raise TypeError(message)

        for key, value in env_vars.items():
            if not isinstance(value, str):
                message = (
                    f"Environment variable `{key}` of field `tool.hatch.envs.{self.name}.env-vars` must be a string"
                )
                raise TypeError(message)

        new_env_vars = {}
        with self.metadata.context.apply_context(self.context):
            for key, value in env_vars.items():
                new_env_vars[key] = self.metadata.context.format(value)

        new_env_vars[AppEnvVars.ENV_ACTIVE] = self.name
        return new_env_vars

    @cached_property
    def env_include(self) -> list[str]:
        """
        ```toml config-example
        [tool.hatch.envs.<ENV_NAME>]
        env-include = [...]
        ```
        """
        env_include = self.config.get("env-include", [])
        if not isinstance(env_include, list):
            message = f"Field `tool.hatch.envs.{self.name}.env-include` must be an array"
            raise TypeError(message)

        for i, pattern in enumerate(env_include, 1):
            if not isinstance(pattern, str):
                message = f"Pattern #{i} of field `tool.hatch.envs.{self.name}.env-include` must be a string"
                raise TypeError(message)

        return ["HATCH_BUILD_*", *env_include] if env_include else env_include

    @cached_property
    def env_exclude(self) -> list[str]:
        """
        ```toml config-example
        [tool.hatch.envs.<ENV_NAME>]
        env-exclude = [...]
        ```
        """
        env_exclude = self.config.get("env-exclude", [])
        if not isinstance(env_exclude, list):
            message = f"Field `tool.hatch.envs.{self.name}.env-exclude` must be an array"
            raise TypeError(message)

        for i, pattern in enumerate(env_exclude, 1):
            if not isinstance(pattern, str):
                message = f"Pattern #{i} of field `tool.hatch.envs.{self.name}.env-exclude` must be a string"
                raise TypeError(message)

        return env_exclude

    @cached_property
    def environment_dependencies_complex(self) -> list[Dependency]:
        from hatch.dep.core import Dependency, InvalidDependencyError

        dependencies_complex: list[Dependency] = []
        with self.apply_context():
            for option in ("dependencies", "extra-dependencies"):
                dependencies = self.config.get(option, [])
                if not isinstance(dependencies, list):
                    message = f"Field `tool.hatch.envs.{self.name}.{option}` must be an array"
                    raise TypeError(message)

                for i, entry in enumerate(dependencies, 1):
                    if not isinstance(entry, str):
                        message = f"Dependency #{i} of field `tool.hatch.envs.{self.name}.{option}` must be a string"
                        raise TypeError(message)

                    try:
                        dependencies_complex.append(Dependency(self.metadata.context.format(entry)))
                    except InvalidDependencyError as e:
                        message = f"Dependency #{i} of field `tool.hatch.envs.{self.name}.{option}` is invalid: {e}"
                        raise ValueError(message) from None

        return dependencies_complex

    @cached_property
    def environment_dependencies(self) -> list[str]:
        """
        The list of all [environment dependencies](../../config/environment/overview.md#dependencies).
        """
        return [str(dependency) for dependency in self.environment_dependencies_complex]

    @cached_property
    def project_dependencies_complex(self) -> list[Dependency]:
        workspace_dependencies = self.workspace.get_dependencies()
        if self.skip_install and not self.features and not self.dependency_groups and not workspace_dependencies:
            return []

        from hatch.dep.core import Dependency
        from hatch.utils.dep import get_complex_dependencies, get_complex_dependency_group, get_complex_features

        all_dependencies_complex = list(map(Dependency, workspace_dependencies))
        dependencies, optional_dependencies = self.app.project.get_dependencies()

        # Format dependencies with context before creating Dependency objects
        with self.apply_context():
            formatted_dependencies = [self.metadata.context.format(dep) for dep in dependencies]
            formatted_optional_dependencies = {
                feature: [self.metadata.context.format(dep) for dep in deps]
                for feature, deps in optional_dependencies.items()
            }

        dependencies_complex = get_complex_dependencies(formatted_dependencies)
        optional_dependencies_complex = get_complex_features(formatted_optional_dependencies)

        if not self.skip_install:
            all_dependencies_complex.extend(dependencies_complex.values())

        for feature in self.features:
            if feature not in optional_dependencies_complex:
                message = (
                    f"Feature `{feature}` of field `tool.hatch.envs.{self.name}.features` is not "
                    f"defined in the dynamic field `project.optional-dependencies`"
                )
                raise ValueError(message)

            all_dependencies_complex.extend([
                dep if isinstance(dep, Dependency) else Dependency(str(dep))
                for dep in optional_dependencies_complex[feature]
            ])

        for dependency_group in self.dependency_groups:
            all_dependencies_complex.extend(
                get_complex_dependency_group(self.app.project.dependency_groups, dependency_group)
            )

        return all_dependencies_complex

    @cached_property
    def project_dependencies(self) -> list[str]:
        """
        The list of all [project dependencies](../../config/metadata.md#dependencies) (if
        [installed](../../config/environment/overview.md#skip-install)), selected
        [optional dependencies](../../config/environment/overview.md#features), and
        workspace dependencies.
        """
        return [str(dependency) for dependency in self.project_dependencies_complex]

    @cached_property
    def local_dependencies_complex(self) -> list[Dependency]:
        from hatch.dep.core import Dependency

        local_dependencies_complex = []
        if not self.skip_install:
            local_dependencies_complex.append(
                Dependency(f"{self.metadata.name} @ {self.root.as_uri()}", editable=self.dev_mode)
            )
        if self.workspace.members:
            local_dependencies_complex.extend(
                Dependency(f"{member.project.metadata.name} @ {member.project.location.as_uri()}", editable=True)
                for member in self.workspace.members
            )

        return local_dependencies_complex

    @cached_property
    def dependencies_complex(self) -> list[Dependency]:
        from hatch.dep.core import Dependency

        all_dependencies_complex = list(self.environment_dependencies_complex)

        # Convert additional_dependencies to Dependency objects
        for dep in self.additional_dependencies:
            if isinstance(dep, Dependency):
                all_dependencies_complex.append(dep)
            else:
                all_dependencies_complex.append(Dependency(str(dep)))

        if self.dependency_groups and not self.skip_install:
            from hatch.utils.dep import get_complex_dependency_group

            for dependency_group in self.dependency_groups:
                all_dependencies_complex.extend(
                    get_complex_dependency_group(self.app.project.dependency_groups, dependency_group)
                )

        if self.builder:
            from hatch.project.constants import BuildEnvVars

            # Convert build requirements to Dependency objects
            for req in self.metadata.build.requires_complex:
                if isinstance(req, Dependency):
                    all_dependencies_complex.append(req)
                else:
                    all_dependencies_complex.append(Dependency(str(req)))

            for target in os.environ.get(BuildEnvVars.REQUESTED_TARGETS, "").split():
                target_config = self.app.project.config.build.target(target)
                all_dependencies_complex.extend(map(Dependency, target_config.dependencies))

            return all_dependencies_complex

        # Ensure these are checked last to speed up initial environment creation since
        # they will already be installed along with the project
        if self.dev_mode or self.features or self.dependency_groups:
            all_dependencies_complex.extend(self.project_dependencies_complex)

        return all_dependencies_complex

    @cached_property
    def dependencies(self) -> list[str]:
        """
        The list of all
        [project dependencies](reference.md#hatch.env.plugin.interface.EnvironmentInterface.project_dependencies)
        (if in [dev mode](../../config/environment/overview.md#dev-mode)) and
        [environment dependencies](../../config/environment/overview.md#dependencies).
        """
        return [str(dependency) for dependency in self.dependencies_complex]

    @cached_property
    def all_dependencies_complex(self) -> list[Dependency]:
        from hatch.dep.core import Dependency

        local_deps = list(self.local_dependencies_complex)
        workspace_names = {dep.name.lower() for dep in local_deps}

        filtered_deps: list[Dependency] = []
        for dep in self.dependencies_complex:
            dep_obj = dep if isinstance(dep, Dependency) else Dependency(str(dep))

            if dep_obj.name.lower() in workspace_names and dep_obj.extras:
                # Only expand if we have static optional dependencies to avoid recursion
                if not self.metadata.hatch.metadata.hook_config:
                    optional_dependencies = self.metadata.core.optional_dependencies
                    for extra in dep_obj.extras:
                        if extra in optional_dependencies:
                            filtered_deps.extend(Dependency(d) for d in optional_dependencies[extra])
            elif dep_obj.name.lower() not in workspace_names:
                filtered_deps.append(dep_obj)

        return local_deps + filtered_deps

    @cached_property
    def all_dependencies(self) -> list[str]:
        return [str(dependency) for dependency in self.all_dependencies_complex]

    @cached_property
    def platforms(self) -> list[str]:
        """
        All names are stored as their lower-cased version.

        ```toml config-example
        [tool.hatch.envs.<ENV_NAME>]
        platforms = [...]
        ```
        """
        platforms = self.config.get("platforms", [])
        if not isinstance(platforms, list):
            message = f"Field `tool.hatch.envs.{self.name}.platforms` must be an array"
            raise TypeError(message)

        for i, command in enumerate(platforms, 1):
            if not isinstance(command, str):
                message = f"Platform #{i} of field `tool.hatch.envs.{self.name}.platforms` must be a string"
                raise TypeError(message)

        return [platform.lower() for platform in platforms]

    @cached_property
    def locked(self) -> bool:
        """
        ```toml config-example
        [tool.hatch.envs.<ENV_NAME>]
        locked = ...
        ```
        """
        global_default = self.app.project.config.lock_envs
        locked = self.config.get("locked", global_default)
        if not isinstance(locked, bool):
            message = f"Field `tool.hatch.envs.{self.name}.locked` must be a boolean"
            raise TypeError(message)

        return locked

    @cached_property
    def skip_install(self) -> bool:
        """
        ```toml config-example
        [tool.hatch.envs.<ENV_NAME>]
        skip-install = ...
        ```
        """
        skip_install = self.config.get("skip-install", not self.metadata.has_project_file())
        if not isinstance(skip_install, bool):
            message = f"Field `tool.hatch.envs.{self.name}.skip-install` must be a boolean"
            raise TypeError(message)

        return skip_install

    @cached_property
    def dev_mode(self) -> bool:
        """
        ```toml config-example
        [tool.hatch.envs.<ENV_NAME>]
        dev-mode = ...
        ```
        """
        dev_mode = self.config.get("dev-mode", True)
        if not isinstance(dev_mode, bool):
            message = f"Field `tool.hatch.envs.{self.name}.dev-mode` must be a boolean"
            raise TypeError(message)

        return dev_mode

    @cached_property
    def builder(self) -> bool:
        """
        ```toml config-example
        [tool.hatch.envs.<ENV_NAME>]
        builder = ...
        ```
        """
        builder = self.config.get("builder", False)
        if not isinstance(builder, bool):
            message = f"Field `tool.hatch.envs.{self.name}.builder` must be a boolean"
            raise TypeError(message)

        return builder

    @cached_property
    def features(self):
        from hatch.utils.metadata import normalize_project_name

        features = self.config.get("features", [])
        if not isinstance(features, list):
            message = f"Field `tool.hatch.envs.{self.name}.features` must be an array of strings"
            raise TypeError(message)

        all_features = set()
        for i, feature in enumerate(features, 1):
            if not isinstance(feature, str):
                message = f"Feature #{i} of field `tool.hatch.envs.{self.name}.features` must be a string"
                raise TypeError(message)

            if not feature:
                message = f"Feature #{i} of field `tool.hatch.envs.{self.name}.features` cannot be an empty string"
                raise ValueError(message)

            normalized_feature = (
                feature if self.metadata.hatch.metadata.allow_ambiguous_features else normalize_project_name(feature)
            )
            if (
                not self.metadata.hatch.metadata.hook_config
                and normalized_feature not in self.metadata.core.optional_dependencies
            ):
                message = (
                    f"Feature `{normalized_feature}` of field `tool.hatch.envs.{self.name}.features` is not "
                    f"defined in field `project.optional-dependencies`"
                )
                raise ValueError(message)

            all_features.add(normalized_feature)

        return sorted(all_features)

    @cached_property
    def dependency_groups(self):
        from hatch.utils.metadata import normalize_project_name

        dependency_groups = self.config.get("dependency-groups", [])
        if not isinstance(dependency_groups, list):
            message = f"Field `tool.hatch.envs.{self.name}.dependency-groups` must be an array of strings"
            raise TypeError(message)

        all_dependency_groups = set()
        for i, dependency_group in enumerate(dependency_groups, 1):
            if not isinstance(dependency_group, str):
                message = (
                    f"Dependency Group #{i} of field `tool.hatch.envs.{self.name}.dependency-groups` must be a string"
                )
                raise TypeError(message)

            if not dependency_group:
                message = f"Dependency Group #{i} of field `tool.hatch.envs.{self.name}.dependency-groups` cannot be an empty string"
                raise ValueError(message)

            normalized_dependency_group = normalize_project_name(dependency_group)
            if (
                not self.metadata.hatch.metadata.hook_config
                and normalized_dependency_group not in self.app.project.dependency_groups
            ):
                message = (
                    f"Dependency Group `{normalized_dependency_group}` of field `tool.hatch.envs.{self.name}.dependency-groups` is not "
                    f"defined in field `dependency-groups`"
                )
                raise ValueError(message)

            all_dependency_groups.add(normalized_dependency_group)

        return sorted(all_dependency_groups)

    @cached_property
    def description(self) -> str:
        """
        ```toml config-example
        [tool.hatch.envs.<ENV_NAME>]
        description = ...
        ```
        """
        description = self.config.get("description", "")
        if not isinstance(description, str):
            message = f"Field `tool.hatch.envs.{self.name}.description` must be a string"
            raise TypeError(message)

        return description

    @cached_property
    def scripts(self):
        config = {}

        # Extra scripts should come first to give less precedence
        for field in ("extra-scripts", "scripts"):
            script_config = self.config.get(field, {})
            if not isinstance(script_config, dict):
                message = f"Field `tool.hatch.envs.{self.name}.{field}` must be a table"
                raise TypeError(message)

            for name, data in script_config.items():
                if " " in name:
                    message = (
                        f"Script name `{name}` in field `tool.hatch.envs.{self.name}.{field}` must not contain spaces"
                    )
                    raise ValueError(message)

                commands = []

                if isinstance(data, str):
                    commands.append(data)
                elif isinstance(data, list):
                    for i, command in enumerate(data, 1):
                        if not isinstance(command, str):
                            message = (
                                f"Command #{i} in field `tool.hatch.envs.{self.name}.{field}.{name}` must be a string"
                            )
                            raise TypeError(message)

                        commands.append(command)
                else:
                    message = (
                        f"Field `tool.hatch.envs.{self.name}.{field}.{name}` must be a string or an array of strings"
                    )
                    raise TypeError(message)

                config[name] = commands

        seen = {}
        active = []
        for script_name, commands in config.items():
            commands[:] = expand_script_commands(self.name, script_name, commands, config, seen, active)

        return config

    @cached_property
    def pre_install_commands(self):
        pre_install_commands = self.config.get("pre-install-commands", [])
        if not isinstance(pre_install_commands, list):
            message = f"Field `tool.hatch.envs.{self.name}.pre-install-commands` must be an array"
            raise TypeError(message)

        for i, command in enumerate(pre_install_commands, 1):
            if not isinstance(command, str):
                message = f"Command #{i} of field `tool.hatch.envs.{self.name}.pre-install-commands` must be a string"
                raise TypeError(message)

        return list(pre_install_commands)

    @cached_property
    def post_install_commands(self):
        post_install_commands = self.config.get("post-install-commands", [])
        if not isinstance(post_install_commands, list):
            message = f"Field `tool.hatch.envs.{self.name}.post-install-commands` must be an array"
            raise TypeError(message)

        for i, command in enumerate(post_install_commands, 1):
            if not isinstance(command, str):
                message = f"Command #{i} of field `tool.hatch.envs.{self.name}.post-install-commands` must be a string"
                raise TypeError(message)

        return list(post_install_commands)

    @cached_property
    def workspace(self) -> Workspace:
        env_config = self.config.get("workspace", {})
        if not isinstance(env_config, dict):
            message = f"Field `tool.hatch.envs.{self.name}.workspace` must be a table"
            raise TypeError(message)

        return Workspace(self, env_config)

    def activate(self):
        """
        A convenience method called when using the environment as a context manager:

        ```python
        with environment:
            ...
        ```
        """

    def deactivate(self):
        """
        A convenience method called after using the environment as a context manager:

        ```python
        with environment:
            ...
        ```
        """

    @abstractmethod
    def find(self):
        """
        :material-align-horizontal-left: **REQUIRED** :material-align-horizontal-right:

        This should return information about how to locate the environment or represent its ID in
        some way. Additionally, this is expected to return something even if the environment is
        [incompatible](reference.md#hatch.env.plugin.interface.EnvironmentInterface.check_compatibility).
        """

    @abstractmethod
    def create(self):
        """
        :material-align-horizontal-left: **REQUIRED** :material-align-horizontal-right:

        This should perform the necessary steps to set up the environment.
        """

    @abstractmethod
    def remove(self):
        """
        :material-align-horizontal-left: **REQUIRED** :material-align-horizontal-right:

        This should perform the necessary steps to completely remove the environment from the system and will only
        be triggered manually by users with the [`env remove`](../../cli/reference.md#hatch-env-remove) or
        [`env prune`](../../cli/reference.md#hatch-env-prune) commands.
        """

    @abstractmethod
    def exists(self) -> bool:
        """
        :material-align-horizontal-left: **REQUIRED** :material-align-horizontal-right:

        This should indicate whether or not the environment has already been created.
        """

    @abstractmethod
    def install_project(self):
        """
        :material-align-horizontal-left: **REQUIRED** :material-align-horizontal-right:

        This should install the project in the environment.
        """

    @abstractmethod
    def install_project_dev_mode(self):
        """
        :material-align-horizontal-left: **REQUIRED** :material-align-horizontal-right:

        This should install the project in the environment such that the environment
        always reflects the current state of the project.
        """

    @abstractmethod
    def dependencies_in_sync(self) -> bool:
        """
        :material-align-horizontal-left: **REQUIRED** :material-align-horizontal-right:

        This should indicate whether or not the environment is compatible with the current
        [dependencies](reference.md#hatch.env.plugin.interface.EnvironmentInterface.dependencies).
        """

    @abstractmethod
    def sync_dependencies(self):
        """
        :material-align-horizontal-left: **REQUIRED** :material-align-horizontal-right:

        This should install the
        [dependencies](reference.md#hatch.env.plugin.interface.EnvironmentInterface.dependencies)
        in the environment.
        """

    def dependency_hash(self):
        """
        This should return a hash of the environment's
        [dependencies](reference.md#hatch.env.plugin.interface.EnvironmentInterface.dependencies)
        and any other data that is handled by the
        [sync_dependencies](reference.md#hatch.env.plugin.interface.EnvironmentInterface.sync_dependencies)
        and
        [dependencies_in_sync](reference.md#hatch.env.plugin.interface.EnvironmentInterface.dependencies_in_sync)
        methods.
        """
        from hatch.utils.dep import hash_dependencies

        return hash_dependencies(self.all_dependencies_complex)

    @contextmanager
    def app_status_creation(self):
        """
        See the [life cycle of environments](reference.md#life-cycle).
        """
        with self.app.status(f"Creating environment: {self.name}"):
            yield

    @contextmanager
    def app_status_pre_installation(self):
        """
        See the [life cycle of environments](reference.md#life-cycle).
        """
        with self.app.status("Running pre-installation commands"):
            yield

    @contextmanager
    def app_status_post_installation(self):
        """
        See the [life cycle of environments](reference.md#life-cycle).
        """
        with self.app.status("Running post-installation commands"):
            yield

    @contextmanager
    def app_status_project_installation(self):
        """
        See the [life cycle of environments](reference.md#life-cycle).
        """
        if self.dev_mode:
            with self.app.status("Installing project in development mode"):
                yield
        else:
            with self.app.status("Installing project"):
                yield

    @contextmanager
    def app_status_dependency_state_check(self):
        """
        See the [life cycle of environments](reference.md#life-cycle).
        """
        if not self.skip_install and (
            "dependencies" in self.metadata.dynamic or "optional-dependencies" in self.metadata.dynamic
        ):
            with self.app.status("Polling dependency state"):
                yield
        else:
            yield

    @contextmanager
    def app_status_dependency_installation_check(self):
        """
        See the [life cycle of environments](reference.md#life-cycle).
        """
        with self.app.status("Checking dependencies"):
            yield

    @contextmanager
    def app_status_dependency_synchronization(self):
        """
        See the [life cycle of environments](reference.md#life-cycle).
        """
        with self.app.status("Syncing dependencies"):
            yield

    @contextmanager
    def fs_context(self) -> Generator[FileSystemContext, None, None]:
        """
        A context manager that must yield a subclass of
        [FileSystemContext](../utilities.md#hatch.env.plugin.interface.FileSystemContext).
        """
        from hatch.utils.fs import temp_directory

        with temp_directory() as temp_dir:
            yield FileSystemContext(self, local_path=temp_dir, env_path=str(temp_dir))

    def enter_shell(
        self,
        name: str,  # noqa: ARG002
        path: str,
        args: Iterable[str],
    ):
        """
        Spawn a [shell](../../config/hatch.md#shell) within the environment.

        This should either use
        [command_context](reference.md#hatch.env.plugin.interface.EnvironmentInterface.command_context)
        directly or provide the same guarantee.
        """
        with self.command_context():
            self.platform.exit_with_command([path, *args])

    def run_shell_command(self, command: str, **kwargs):
        """
        This should return the standard library's
        [subprocess.CompletedProcess](https://docs.python.org/3/library/subprocess.html#subprocess.CompletedProcess)
        and will always be called when the
        [command_context](reference.md#hatch.env.plugin.interface.EnvironmentInterface.command_context)
        is active, with the expectation of providing the same guarantee.
        """
        kwargs.setdefault("shell", True)
        return self.platform.run_command(command, **kwargs)

    @contextmanager
    def command_context(self):
        """
        A context manager that when active should make executed shell commands reflect any
        [environment variables](reference.md#hatch.env.plugin.interface.EnvironmentInterface.get_env_vars)
        the user defined either currently or at the time of
        [creation](reference.md#hatch.env.plugin.interface.EnvironmentInterface.create).

        For an example, open the default implementation below:
        """
        with self.get_env_vars():
            yield

    def resolve_commands(self, commands: list[str]):
        """
        This expands each command into one or more commands based on any
        [scripts](../../config/environment/overview.md#scripts) that the user defined.
        """
        for command in commands:
            yield from self.expand_command(command)

    def expand_command(self, command):
        possible_script, args, _ignore_exit_code = parse_script_command(command)

        # Indicate undefined
        if not args:
            args = None

        with self.apply_context():
            if possible_script in self.scripts:
                if args is not None:
                    args = self.metadata.context.format(args)

                for cmd in self.scripts[possible_script]:
                    yield self.metadata.context.format(cmd, args=args).strip()
            else:
                yield self.metadata.context.format(command, args=args).strip()

    def construct_pip_install_command(self, args: list[str]):
        """
        A convenience method for constructing a [`pip install`](https://pip.pypa.io/en/stable/cli/pip_install/)
        command with the given verbosity. The default verbosity is set to one less than Hatch's verbosity.
        """
        command = ["python", "-u", "-m", "pip", "install", "--disable-pip-version-check"]

        # Default to -1 verbosity
        add_verbosity_flag(command, self.verbosity, adjustment=-1)

        command.extend(args)
        return command

    def join_command_args(self, args: list[str]):
        """
        This is used by the [`run`](../../cli/reference.md#hatch-run) command to construct the root command string
        from the received arguments.
        """
        return self.platform.join_command_args(args)

    def apply_features(self, requirement: str):
        """
        A convenience method that applies any user defined [features](../../config/environment/overview.md#features)
        to the given requirement.
        """
        if self.features:
            features = ",".join(self.features)
            return f"{requirement}[{features}]"

        return requirement

    def check_compatibility(self):
        """
        This raises an exception if the environment is not compatible with the user's setup. The default behavior
        checks for [platform compatibility](../../config/environment/overview.md#supported-platforms)
        and any method override should keep this check.

        This check is never performed if the environment has been
        [created](reference.md#hatch.env.plugin.interface.EnvironmentInterface.create).
        """
        if self.platforms and self.platform.name not in self.platforms:
            message = "unsupported platform"
            raise OSError(message)

    def get_env_vars(self) -> EnvVars:
        """
        Returns a mapping of environment variables that should be available to the environment. The object can
        be used as a context manager to temporarily apply the environment variables to the current process.

        !!! note
            The environment variable `HATCH_ENV_ACTIVE` will always be set to the name of the environment.
        """
        return EnvVars(self.env_vars, self.env_include, self.env_exclude)

    def get_env_var_option(self, option: str) -> str:
        """
        Returns the value of the upper-cased environment variable `HATCH_ENV_TYPE_<PLUGIN_NAME>_<option>`.
        """
        return get_env_var_option(plugin_name=self.PLUGIN_NAME, option=option)

    def get_context(self):
        """
        Returns a subclass of
        [EnvironmentContextFormatter](../utilities.md#hatch.env.context.EnvironmentContextFormatter).
        """
        from hatch.env.context import EnvironmentContextFormatter

        return EnvironmentContextFormatter(self)

    @staticmethod
    def get_option_types() -> dict:
        """
        Returns a mapping of supported options to their respective types so that they can be used by
        [overrides](../../config/environment/advanced.md#option-overrides).
        """
        return {}

    @contextmanager
    def apply_context(self):
        with self.get_env_vars(), self.metadata.context.apply_context(self.context):
            yield

    def __enter__(self):
        self.activate()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.deactivate()


class FileSystemContext:
    """
    This class represents a synchronized path between the local file system and a potentially remote environment.
    """

    def __init__(self, env: EnvironmentInterface, *, local_path: Path, env_path: str):
        self.__env = env
        self.__local_path = local_path
        self.__env_path = env_path

    @property
    def env(self) -> EnvironmentInterface:
        """
        Returns the environment to which this context belongs.
        """
        return self.__env

    @property
    def local_path(self) -> Path:
        """
        Returns the local path to which this context refers as a path-like object.
        """
        return self.__local_path

    @property
    def env_path(self) -> str:
        """
        Returns the environment path to which this context refers as a string. The environment
        may not be on the local file system.
        """
        return self.__env_path

    def join(self, relative_path: str) -> FileSystemContext:
        """
        Returns a new instance of this class with the given relative path appended to the local and
        environment paths.

        This method should not need overwriting.
        """
        local_path = self.local_path / relative_path
        env_path = f"{self.env_path}{self.__env.sep.join(['', *os.path.normpath(relative_path).split(os.sep)])}"
        return FileSystemContext(self.__env, local_path=local_path, env_path=env_path)

    def sync_env(self):
        """
        Synchronizes the [environment path](utilities.md#hatch.env.plugin.interface.FileSystemContext.env_path)
        with the [local path](utilities.md#hatch.env.plugin.interface.FileSystemContext.local_path) as the source.
        """

    def sync_local(self):
        """
        Synchronizes the [local path](utilities.md#hatch.env.plugin.interface.FileSystemContext.local_path)
        with the [environment path](utilities.md#hatch.env.plugin.interface.FileSystemContext.env_path) as
        the source.
        """


class Workspace:
    def __init__(self, env: EnvironmentInterface, config: dict[str, Any]):
        self.env = env
        self.config = config

    @cached_property
    def parallel(self) -> bool:
        parallel = self.config.get("parallel", True)
        if not isinstance(parallel, bool):
            message = f"Field `tool.hatch.envs.{self.env.name}.workspace.parallel` must be a boolean"
            raise TypeError(message)

        return parallel

    def get_dependencies(self) -> list[str]:
        static_members: list[WorkspaceMember] = []
        dynamic_members: list[WorkspaceMember] = []
        for member in self.members:
            if member.has_static_dependencies:
                static_members.append(member)
            else:
                dynamic_members.append(member)

        all_dependencies = []
        for member in static_members:
            dependencies, features = member.get_dependencies()
            all_dependencies.extend(dependencies)
            for feature in member.features:
                all_dependencies.extend(features.get(feature, []))

        if self.parallel:
            from concurrent.futures import ThreadPoolExecutor

            def get_member_deps(member):
                with self.env.app.status(f"Checking workspace member: {member.name}"):
                    dependencies, features = member.get_dependencies()
                    deps = list(dependencies)
                    for feature in member.features:
                        deps.extend(features.get(feature, []))
                    return deps

            with ThreadPoolExecutor() as executor:
                results = executor.map(get_member_deps, dynamic_members)
                for deps in results:
                    all_dependencies.extend(deps)
        else:
            for member in dynamic_members:
                with self.env.app.status(f"Checking workspace member: {member.name}"):
                    dependencies, features = member.get_dependencies()
                    all_dependencies.extend(dependencies)
                    for feature in member.features:
                        all_dependencies.extend(features.get(feature, []))

        return all_dependencies

    @cached_property
    def members(self) -> list[WorkspaceMember]:
        import fnmatch

        from hatch.project.core import Project
        from hatch.utils.fs import Path
        from hatch.utils.metadata import normalize_project_name

        raw_members = self.config.get("members", [])
        if not isinstance(raw_members, list):
            message = f"Field `tool.hatch.envs.{self.env.name}.workspace.members` must be an array"
            raise TypeError(message)

        # Get exclude patterns
        exclude_patterns = self.config.get("exclude", [])
        if not isinstance(exclude_patterns, list):
            message = f"Field `tool.hatch.envs.{self.env.name}.workspace.exclude` must be an array"
            raise TypeError(message)

        # First normalize configuration with context expansion
        member_data: list[dict[str, Any]] = []
        with self.env.apply_context():
            for i, data in enumerate(raw_members, 1):
                if isinstance(data, str):
                    expanded_path = self.env.metadata.context.format(data)
                    member_data.append({"path": expanded_path, "features": ()})
                elif isinstance(data, dict):
                    if "path" not in data:
                        message = (
                            f"Member #{i} of field `tool.hatch.envs.{self.env.name}.workspace.members` must define "
                            f"a `path` key"
                        )
                        raise TypeError(message)

                    path = data["path"]
                    if not isinstance(path, str):
                        message = (
                            f"Option `path` of member #{i} of field `tool.hatch.envs.{self.env.name}.workspace.members` "
                            f"must be a string"
                        )
                        raise TypeError(message)

                    if not path:
                        message = (
                            f"Option `path` of member #{i} of field `tool.hatch.envs.{self.env.name}.workspace.members` "
                            f"cannot be an empty string"
                        )
                        raise ValueError(message)

                    expanded_path = self.env.metadata.context.format(path)

                    features = data.get("features", [])
                    if not isinstance(features, list):
                        message = (
                            f"Option `features` of member #{i} of field `tool.hatch.envs.{self.env.name}.workspace."
                            f"members` must be an array of strings"
                        )
                        raise TypeError(message)

                    all_features: set[str] = set()
                    for j, feature in enumerate(features, 1):
                        if not isinstance(feature, str):
                            message = (
                                f"Feature #{j} of option `features` of member #{i} of field "
                                f"`tool.hatch.envs.{self.env.name}.workspace.members` must be a string"
                            )
                            raise TypeError(message)

                        if not feature:
                            message = (
                                f"Feature #{j} of option `features` of member #{i} of field "
                                f"`tool.hatch.envs.{self.env.name}.workspace.members` cannot be an empty string"
                            )
                            raise ValueError(message)

                        normalized_feature = normalize_project_name(feature)
                        if normalized_feature in all_features:
                            message = (
                                f"Feature #{j} of option `features` of member #{i} of field "
                                f"`tool.hatch.envs.{self.env.name}.workspace.members` is a duplicate"
                            )
                            raise ValueError(message)

                        all_features.add(normalized_feature)

                    member_data.append({"path": expanded_path, "features": tuple(sorted(all_features))})
                else:
                    message = (
                        f"Member #{i} of field `tool.hatch.envs.{self.env.name}.workspace.members` must be "
                        f"a string or an inline table"
                    )
                    raise TypeError(message)

        root = str(self.env.root)
        member_paths: dict[str, WorkspaceMember] = {}
        for data in member_data:
            # Given root R and member spec M, we need to find:
            #
            # 1. The absolute path AP of R/M
            # 2. The shared prefix SP of R and AP
            # 3. The relative path RP of M from AP
            #
            # For example, if:
            #
            # R = /foo/bar/baz
            # M = ../dir/pkg-*
            #
            # Then:
            #
            # AP = /foo/bar/dir/pkg-*
            # SP = /foo/bar
            # RP = dir/pkg-*
            path_spec = data["path"]
            normalized_path = os.path.normpath(os.path.join(root, path_spec))
            absolute_path = os.path.abspath(normalized_path)
            shared_prefix = os.path.commonpath([root, absolute_path])
            relative_path = os.path.relpath(absolute_path, shared_prefix)

            # Now we have the necessary information to perform an optimized glob search for members
            members_found = False
            for member_path in find_members(shared_prefix, relative_path.split(os.sep)):
                # Check if member should be excluded
                relative_member_path = os.path.relpath(member_path, shared_prefix)
                should_exclude = False
                for exclude_pattern in exclude_patterns:
                    if fnmatch.fnmatch(relative_member_path, exclude_pattern) or fnmatch.fnmatch(
                        member_path, exclude_pattern
                    ):
                        should_exclude = True
                        break

                if should_exclude:
                    continue

                project_file = os.path.join(member_path, "pyproject.toml")
                if not os.path.isfile(project_file):
                    message = (
                        f"Member derived from `{path_spec}` of field "
                        f"`tool.hatch.envs.{self.env.name}.workspace.members` is not a project (no `pyproject.toml` "
                        f"file): {member_path}"
                    )
                    raise OSError(message)

                members_found = True
                if member_path in member_paths:
                    message = (
                        f"Member derived from `{path_spec}` of field "
                        f"`tool.hatch.envs.{self.env.name}.workspace.members` is a duplicate: {member_path}"
                    )
                    raise ValueError(message)

                project = Project(Path(member_path), locate=False)
                project.set_app(self.env.app)
                member_paths[member_path] = WorkspaceMember(project, features=data["features"])

            if not members_found:
                message = (
                    f"No members could be derived from `{path_spec}` of field "
                    f"`tool.hatch.envs.{self.env.name}.workspace.members`: {absolute_path}"
                )
                raise OSError(message)

        return list(member_paths.values())


class WorkspaceMember:
    def __init__(self, project: Project, *, features: tuple[str]):
        self.project = project
        self.features = features
        self._last_modified: float

    @cached_property
    def name(self) -> str:
        return self.project.metadata.name

    @cached_property
    def has_static_dependencies(self) -> bool:
        return self.project.has_static_dependencies

    def get_dependencies(self) -> tuple[list[str], dict[str, list[str]]]:
        return self.project.get_dependencies()

    @property
    def last_modified(self) -> float:
        """Get the last modification time of the member's pyproject.toml."""
        import os

        pyproject_path = self.project.location / "pyproject.toml"
        if pyproject_path.exists():
            return os.path.getmtime(pyproject_path)
        return 0.0

    def get_editable_requirement(self, *, editable: bool = True) -> str:
        """Get the requirement string for this workspace member."""
        uri = self.project.location.as_uri()
        if editable:
            return f"-e {self.name} @ {uri}"
        return f"{self.name} @ {uri}"


def expand_script_commands(env_name, script_name, commands, config, seen, active):
    if script_name in seen:
        return seen[script_name]

    if script_name in active:
        active.append(script_name)

        message = f"Circular expansion detected for field `tool.hatch.envs.{env_name}.scripts`: {' -> '.join(active)}"
        raise ValueError(message)

    active.append(script_name)

    expanded_commands = []

    for command in commands:
        possible_script, args, ignore_exit_code = parse_script_command(command)

        if possible_script in config:
            expanded_commands.extend(
                format_script_commands(
                    commands=expand_script_commands(
                        env_name, possible_script, config[possible_script], config, seen, active
                    ),
                    args=args,
                    ignore_exit_code=ignore_exit_code,
                )
            )
        else:
            expanded_commands.append(command)

    seen[script_name] = expanded_commands
    active.pop()

    return expanded_commands


def find_members(root, relative_components):
    import fnmatch
    import re

    component_matchers = []
    for component in relative_components:
        if any(special in component for special in "*?["):
            pattern = re.compile(fnmatch.translate(component))
            component_matchers.append(lambda entry, pattern=pattern: pattern.search(entry.name))
        else:
            component_matchers.append(lambda entry, component=component: component == entry.name)

    results = list(_recurse_members(root, 0, component_matchers))
    yield from sorted(results, key=os.path.basename)


def _recurse_members(root, matcher_index, matchers):
    if matcher_index == len(matchers):
        yield root
        return

    matcher = matchers[matcher_index]
    with os.scandir(root) as it:
        for entry in it:
            if entry.is_dir() and matcher(entry):
                yield from _recurse_members(entry.path, matcher_index + 1, matchers)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Locate Python binaries")
@click.argument("name")
@click.option("-p", "--parent", is_flag=True, help="Show the parent directory of the Python binary")
@click.option("--dir", "-d", "directory", help="The directory in which distributions reside")
@click.pass_obj
def find(app: Application, *, name: str, parent: bool, directory: str | None):
    """Locate Python binaries."""
    manager = app.get_python_manager(directory)
    installed = manager.get_installed()
    if name not in installed:
        app.abort(f"Distribution not installed: {name}")

    dist = installed[name]
    app.display(str(dist.python_path.parent if parent else dist.python_path))
